"""Published Condition data type for the Data Store Client."""

from __future__ import annotations

from ni.datastore.data._types._moniker import Moniker
from ni.measurements.data.v1.data_store_pb2 import (
    PublishedCondition as PublishedConditionProto,
)


class PublishedCondition:
    """Represents a condition that has been published to the data store.

    A published condition contains metadata about a condition value that was
    published, including a moniker for data retrieval and associated metadata
    like condition name, type, and associated step/test result IDs.
    """

    __slots__ = (
        "moniker",
        "id",
        "name",
        "condition_type",
        "step_id",
        "test_result_id",
    )

    def __init__(
        self,
        *,
        moniker: Moniker | None = None,
        id: str = "",
        name: str = "",
        condition_type: str = "",
        step_id: str = "",
        test_result_id: str = "",
    ) -> None:
        """Initialize a PublishedCondition instance.

        Args:
            moniker: The moniker of the condition that this value is associated
                with. This moniker returns a Vector when read.
            id: The unique identifier of the condition. This
                can be used to reference and find the condition in the data store.
            name: The name of the condition.
            condition_type: The type of the condition. For example, "Setup" or
                "Environment".
            step_id: The ID of the step with which this condition is associated.
            test_result_id: The ID of the test result with which this condition
                is associated.
        """
        self.moniker = moniker
        self.id = id
        self.name = name
        self.condition_type = condition_type
        self.step_id = step_id
        self.test_result_id = test_result_id

    @staticmethod
    def from_protobuf(published_condition_proto: PublishedConditionProto) -> "PublishedCondition":
        """Create a PublishedCondition instance from a protobuf PublishedCondition message."""
        return PublishedCondition(
            moniker=(
                Moniker.from_protobuf(published_condition_proto.moniker)
                if published_condition_proto.HasField("moniker")
                else None
            ),
            id=published_condition_proto.id,
            name=published_condition_proto.name,
            condition_type=published_condition_proto.condition_type,
            step_id=published_condition_proto.step_id,
            test_result_id=published_condition_proto.test_result_id,
        )

    def to_protobuf(self) -> PublishedConditionProto:
        """Convert this PublishedCondition instance to a protobuf PublishedCondition message."""
        return PublishedConditionProto(
            moniker=self.moniker.to_protobuf() if self.moniker is not None else None,
            id=self.id,
            name=self.name,
            condition_type=self.condition_type,
            step_id=self.step_id,
            test_result_id=self.test_result_id,
        )

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, PublishedCondition):
            return NotImplemented
        return (
            self.moniker == other.moniker
            and self.id == other.id
            and self.name == other.name
            and self.condition_type == other.condition_type
            and self.step_id == other.step_id
            and self.test_result_id == other.test_result_id
        )

    def __str__(self) -> str:
        """Return a string representation of the PublishedCondition."""
        return str(self.to_protobuf())



"""Alias data type for the Data Store Client."""

from __future__ import annotations

from ni.measurements.metadata.v1.metadata_store_pb2 import Alias as AliasProto

from ._alias_target_type import AliasTargetType


class Alias:
    """Represents an alias for metadata.

    An alias provides a user-friendly name that maps to a specific metadata
    instance, containing the alias name, the type of the aliased metadata
    instance, and the unique identifier for the aliased metadata instance.
    """

    __slots__ = (
        "name",
        "target_type",
        "target_id",
    )

    def __init__(
        self,
        *,
        name: str = "",
        target_type: AliasTargetType = AliasTargetType.UNSPECIFIED,
        target_id: str = "",
    ) -> None:
        """Initialize an Alias instance.

        Args:
            name: The registered alias name for the aliased metadata
                instance.
            target_type: The type of the aliased metadata instance.
            target_id: The unique identifier for the aliased metadata instance.
        """
        self.name = name
        self.target_type = target_type
        self.target_id = target_id

    @staticmethod
    def from_protobuf(alias_proto: AliasProto) -> "Alias":
        """Create an Alias instance from a protobuf Alias message."""
        return Alias(
            name=alias_proto.name,
            target_type=AliasTargetType.from_protobuf(alias_proto.target_type),
            target_id=alias_proto.target_id,
        )

    def to_protobuf(self) -> AliasProto:
        """Convert this Alias to a protobuf Alias message."""
        return AliasProto(
            name=self.name,
            target_type=self.target_type.to_protobuf(),
            target_id=self.target_id,
        )

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, Alias):
            return NotImplemented
        return (
            self.name == other.name
            and self.target_type == other.target_type
            and self.target_id == other.target_id
        )

    def __str__(self) -> str:
        """Return a string representation of the Alias."""
        return str(self.to_protobuf())



"""Metadata store client for publishing and reading metadata."""

from __future__ import annotations

import logging
import sys
from collections.abc import Sequence
from pathlib import Path
from threading import Lock
from types import TracebackType
from typing import TYPE_CHECKING

from grpc import Channel
from ni.datastore.metadata._types._alias import Alias
from ni.datastore.metadata._types._extension_schema import ExtensionSchema
from ni.datastore.metadata._types._hardware_item import HardwareItem
from ni.datastore.metadata._types._metadata_items import MetadataItems
from ni.datastore.metadata._types._operator import Operator
from ni.datastore.metadata._types._software_item import SoftwareItem
from ni.datastore.metadata._types._test import Test
from ni.datastore.metadata._types._test_adapter import TestAdapter
from ni.datastore.metadata._types._test_description import TestDescription
from ni.datastore.metadata._types._test_station import TestStation
from ni.datastore.metadata._types._uut import Uut
from ni.datastore.metadata._types._uut_instance import UutInstance
from ni.measurementlink.discovery.v1.client import DiscoveryClient
from ni.measurements.metadata.v1.client import (
    MetadataStoreClient as MetadataStoreServiceClient,
)
from ni.measurements.metadata.v1.metadata_store_service_pb2 import (
    CreateAliasRequest,
    CreateFromJsonDocumentRequest,
    CreateHardwareItemRequest,
    CreateOperatorRequest,
    CreateSoftwareItemRequest,
    CreateTestAdapterRequest,
    CreateTestDescriptionRequest,
    CreateTestRequest,
    CreateTestStationRequest,
    CreateUutInstanceRequest,
    CreateUutRequest,
    DeleteAliasRequest,
    GetAliasRequest,
    GetHardwareItemRequest,
    GetOperatorRequest,
    GetSoftwareItemRequest,
    GetTestAdapterRequest,
    GetTestDescriptionRequest,
    GetTestRequest,
    GetTestStationRequest,
    GetUutInstanceRequest,
    GetUutRequest,
    ListSchemasRequest,
    QueryAliasesRequest,
    QueryHardwareItemsRequest,
    QueryOperatorsRequest,
    QuerySoftwareItemsRequest,
    QueryTestAdaptersRequest,
    QueryTestDescriptionsRequest,
    QueryTestsRequest,
    QueryTestStationsRequest,
    QueryUutInstancesRequest,
    QueryUutsRequest,
    RegisterSchemaRequest,
)
from ni_grpc_extensions.channelpool import GrpcChannelPool

if TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

_logger = logging.getLogger(__name__)


class MetadataStoreClient:
    """Metadata store client for publishing and reading metadata."""

    __slots__ = (
        "_closed",
        "_discovery_client",
        "_grpc_channel",
        "_grpc_channel_pool",
        "_metadata_store_client",
        "_metadata_store_client_lock",
    )

    _METADATA_STORE_CLIENT_CLOSED_ERROR = (
        "This MetadataStoreClient has been closed. Create a new MetadataStoreClient for "
        "further interaction with the metadata store."
    )

    _closed: bool
    _discovery_client: DiscoveryClient | None
    _grpc_channel: Channel | None
    _grpc_channel_pool: GrpcChannelPool | None
    _metadata_store_client: MetadataStoreServiceClient | None
    _metadata_store_client_lock: Lock

    def __init__(
        self,
        discovery_client: DiscoveryClient | None = None,
        grpc_channel: Channel | None = None,
        grpc_channel_pool: GrpcChannelPool | None = None,
    ) -> None:
        """Initialize the MetadataStoreClient.

        Args:
            discovery_client: An optional discovery client (recommended).

            grpc_channel: An optional metadata store gRPC channel. Providing this channel
                will bypass discovery service resolution of the metadata store.

            grpc_channel_pool: An optional gRPC channel pool (recommended).
        """
        self._discovery_client = discovery_client
        self._grpc_channel = grpc_channel
        self._grpc_channel_pool = grpc_channel_pool

        self._metadata_store_client = None
        self._metadata_store_client_lock = Lock()

        self._closed = False

    def __enter__(self) -> Self:
        """Enter the runtime context of the metadata store client."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit the runtime context of the metadata store client."""
        self.close()

    def close(self) -> None:
        """Close the metadata store client and clean up resources that it owns."""
        self._closed = True

        with self._metadata_store_client_lock:
            if self._metadata_store_client is not None:
                self._metadata_store_client.close()
                self._metadata_store_client = None

    def create_uut_instance(self, uut_instance: UutInstance) -> str:
        """Create a new UUT instance in the metadata store.

        Args:
            uut_instance: The metadata of the UUT instance to be created.

        Returns:
            str: The identifier of the created UUT instance.
        """
        create_request = CreateUutInstanceRequest(uut_instance=uut_instance.to_protobuf())
        create_response = self._get_metadata_store_client().create_uut_instance(create_request)
        return create_response.uut_instance_id

    def get_uut_instance(self, uut_instance_id: str) -> UutInstance:
        """Get the UUT instance associated with the given identifier.

        Args:
            uut_instance_id: The identifier of the desired UUT instance.

        Returns:
            UutInstance: The metadata of the requested UUT instance.
        """
        get_request = GetUutInstanceRequest(uut_instance_id=uut_instance_id)
        get_response = self._get_metadata_store_client().get_uut_instance(get_request)
        return UutInstance.from_protobuf(get_response.uut_instance)

    def query_uut_instances(self, odata_query: str = "") -> Sequence[UutInstance]:
        """Perform an OData query on UUT instances.

        Args:
            odata_query: An OData query string. Example: "$filter=name eq
                'Value'". $expand is not supported.

        Returns:
            Sequence[UutInstance]: The list of UUT instances that match the
                query.
        """
        query_request = QueryUutInstancesRequest(odata_query=odata_query)
        query_response = self._get_metadata_store_client().query_uut_instances(query_request)
        return [
            UutInstance.from_protobuf(uut_instance) for uut_instance in query_response.uut_instances
        ]

    def create_uut(self, uut: Uut) -> str:
        """Create a new UUT in the metadata store.

        Args:
            uut: The metadata of the UUT to be created.

        Returns:
            str: The identifier of the created UUT.
        """
        create_request = CreateUutRequest(uut=uut.to_protobuf())
        create_response = self._get_metadata_store_client().create_uut(create_request)
        return create_response.uut_id

    def get_uut(self, uut_id: str) -> Uut:
        """Get the UUT associated with the given identifier.

        Args:
            uut_id: The identifier of the desired UUT.

        Returns:
            Uut: The metadata of the requested UUT.
        """
        get_request = GetUutRequest(uut_id=uut_id)
        get_response = self._get_metadata_store_client().get_uut(get_request)
        return Uut.from_protobuf(get_response.uut)

    def query_uuts(self, odata_query: str = "") -> Sequence[Uut]:
        """Perform an OData query on UUTs.

        Args:
            odata_query: An OData query string. Example: "$filter=name eq
                'Value'". $expand is not supported.

        Returns:
            Sequence[Uut]: The list of UUTs that match the query.
        """
        query_request = QueryUutsRequest(odata_query=odata_query)
        query_response = self._get_metadata_store_client().query_uuts(query_request)
        return [Uut.from_protobuf(uut) for uut in query_response.uuts]

    def create_operator(self, operator: Operator) -> str:
        """Create a new operator in the metadata store.

        Args:
            operator: The metadata of the operator to be created.

        Returns:
            str: The identifier of the created operator.
        """
        create_request = CreateOperatorRequest(operator=operator.to_protobuf())
        create_response = self._get_metadata_store_client().create_operator(create_request)
        return create_response.operator_id

    def get_operator(self, operator_id: str) -> Operator:
        """Get the operator associated with the given identifier.

        Args:
            operator_id: The identifier of the desired operator.

        Returns:
            Operator: The metadata of the requested operator.
        """
        get_request = GetOperatorRequest(operator_id=operator_id)
        get_response = self._get_metadata_store_client().get_operator(get_request)
        return Operator.from_protobuf(get_response.operator)

    def query_operators(self, odata_query: str = "") -> Sequence[Operator]:
        """Perform an OData query on operators.

        Args:
            odata_query: An OData query string. Example: "$filter=name eq
                'Value'". $expand is not supported.

        Returns:
            Sequence[Operator]: The list of operators that match the query.
        """
        query_request = QueryOperatorsRequest(odata_query=odata_query)
        query_response = self._get_metadata_store_client().query_operators(query_request)
        return [Operator.from_protobuf(operator) for operator in query_response.operators]

    def create_test_description(self, test_description: TestDescription) -> str:
        """Create a test description in the metadata store.

        Args:
            test_description: The metadata of the test description to be created.

        Returns:
            str: The identifier of the created test description.
        """
        create_request = CreateTestDescriptionRequest(
            test_description=test_description.to_protobuf()
        )
        create_response = self._get_metadata_store_client().create_test_description(create_request)
        return create_response.test_description_id

    def get_test_description(self, test_description_id: str) -> TestDescription:
        """Get a test description from the metadata store.

        Args:
            test_description_id: The identifier of the desired test description.

        Returns:
            TestDescription: The metadata of the requested test description.
        """
        get_request = GetTestDescriptionRequest(test_description_id=test_description_id)
        get_response = self._get_metadata_store_client().get_test_description(get_request)
        return TestDescription.from_protobuf(get_response.test_description)

    def query_test_descriptions(self, odata_query: str = "") -> Sequence[TestDescription]:
        """Query test descriptions from the metadata store.

        Args:
            odata_query: An OData query string. Example: "$filter=name eq
                'Value'". $expand is not supported.

        Returns:
            Sequence[TestDescription]: The list of test descriptions that match
                the query.
        """
        query_request = QueryTestDescriptionsRequest(odata_query=odata_query)
        query_response = self._get_metadata_store_client().query_test_descriptions(query_request)
        return [
            TestDescription.from_protobuf(test_description)
            for test_description in query_response.test_descriptions
        ]

    def create_test(self, test: Test) -> str:
        """Create a test in the metadata store.

        Args:
            test: The metadata of the test to be created.

        Returns:
            str: The identifier of the created test.
        """
        create_request = CreateTestRequest(test=test.to_protobuf())
        create_response = self._get_metadata_store_client().create_test(create_request)
        return create_response.test_id

    def get_test(self, test_id: str) -> Test:
        """Get a test from the metadata store.

        Args:
            test_id: The identifier of the desired test.

        Returns:
            Test: The metadata of the requested test.
        """
        get_request = GetTestRequest(test_id=test_id)
        get_response = self._get_metadata_store_client().get_test(get_request)
        return Test.from_protobuf(get_response.test)

    def query_tests(self, odata_query: str = "") -> Sequence[Test]:
        """Query tests from the metadata store."""
        query_request = QueryTestsRequest(odata_query=odata_query)
        query_response = self._get_metadata_store_client().query_tests(query_request)
        return [Test.from_protobuf(test) for test in query_response.tests]

    def create_test_station(self, test_station: TestStation) -> str:
        """Create a test station in the metadata store.

        Args:
            test_station: The metadata of the test station to be created.

        Returns:
            str: The identifier of the created test station.
        """
        create_request = CreateTestStationRequest(test_station=test_station.to_protobuf())
        create_response = self._get_metadata_store_client().create_test_station(create_request)
        return create_response.test_station_id

    def get_test_station(self, test_station_id: str) -> TestStation:
        """Get a test station from the metadata store.

        Args:
            test_station_id: The identifier of the desired test station.

        Returns:
            TestStation: The metadata of the requested test station.
        """
        get_request = GetTestStationRequest(test_station_id=test_station_id)
        get_response = self._get_metadata_store_client().get_test_station(get_request)
        return TestStation.from_protobuf(get_response.test_station)

    def query_test_stations(self, odata_query: str = "") -> Sequence[TestStation]:
        """Query test stations from the metadata store."""
        query_request = QueryTestStationsRequest(odata_query=odata_query)
        query_response = self._get_metadata_store_client().query_test_stations(query_request)
        return [
            TestStation.from_protobuf(test_station) for test_station in query_response.test_stations
        ]

    def create_hardware_item(self, hardware_item: HardwareItem) -> str:
        """Create a hardware item in the metadata store.

        Args:
            hardware_item: The metadata of the hardware item to be created.

        Returns:
            str: The identifier of the created hardware item.
        """
        create_request = CreateHardwareItemRequest(hardware_item=hardware_item.to_protobuf())
        create_response = self._get_metadata_store_client().create_hardware_item(create_request)
        return create_response.hardware_item_id

    def get_hardware_item(self, hardware_item_id: str) -> HardwareItem:
        """Get a hardware item from the metadata store.

        Args:
            hardware_item_id: The identifier of the desired hardware item.

        Returns:
            HardwareItem: The metadata of the requested hardware item.
        """
        get_request = GetHardwareItemRequest(hardware_item_id=hardware_item_id)
        get_response = self._get_metadata_store_client().get_hardware_item(get_request)
        return HardwareItem.from_protobuf(get_response.hardware_item)

    def query_hardware_items(self, odata_query: str = "") -> Sequence[HardwareItem]:
        """Query hardware items from the metadata store."""
        query_request = QueryHardwareItemsRequest(odata_query=odata_query)
        query_response = self._get_metadata_store_client().query_hardware_items(query_request)
        return [
            HardwareItem.from_protobuf(hardware_item)
            for hardware_item in query_response.hardware_items
        ]

    def create_software_item(self, software_item: SoftwareItem) -> str:
        """Create a software item in the metadata store.

        Args:
            software_item: The metadata of the software item to be created.

        Returns:
            str: The identifier of the created software item.
        """
        create_request = CreateSoftwareItemRequest(software_item=software_item.to_protobuf())
        create_response = self._get_metadata_store_client().create_software_item(create_request)
        return create_response.software_item_id

    def get_software_item(self, software_item_id: str) -> SoftwareItem:
        """Get a software item from the metadata store.

        Args:
            software_item_id: The identifier of the desired software item.

        Returns:
            SoftwareItem: The metadata of the requested software item.
        """
        get_request = GetSoftwareItemRequest(software_item_id=software_item_id)
        get_response = self._get_metadata_store_client().get_software_item(get_request)
        return SoftwareItem.from_protobuf(get_response.software_item)

    def query_software_items(self, odata_query: str = "") -> Sequence[SoftwareItem]:
        """Query software items from the metadata store."""
        query_request = QuerySoftwareItemsRequest(odata_query=odata_query)
        query_response = self._get_metadata_store_client().query_software_items(query_request)
        return [
            SoftwareItem.from_protobuf(software_item)
            for software_item in query_response.software_items
        ]

    def create_test_adapter(self, test_adapter: TestAdapter) -> str:
        """Create a test adapter in the metadata store.

        Args:
            test_adapter: The metadata of the test adapter to be created.

        Returns:
            str: The identifier of the created test adapter.
        """
        create_request = CreateTestAdapterRequest(test_adapter=test_adapter.to_protobuf())
        create_response = self._get_metadata_store_client().create_test_adapter(create_request)
        return create_response.test_adapter_id

    def get_test_adapter(self, test_adapter_id: str) -> TestAdapter:
        """Get a test adapter from the metadata store.

        Args:
            test_adapter_id: The identifier of the desired test adapter.

        Returns:
            TestAdapter: The metadata of the requested test adapter.
        """
        get_request = GetTestAdapterRequest(test_adapter_id=test_adapter_id)
        get_response = self._get_metadata_store_client().get_test_adapter(get_request)
        return TestAdapter.from_protobuf(get_response.test_adapter)

    def query_test_adapters(self, odata_query: str = "") -> Sequence[TestAdapter]:
        """Query test adapters from the metadata store."""
        query_request = QueryTestAdaptersRequest(odata_query=odata_query)
        query_response = self._get_metadata_store_client().query_test_adapters(query_request)
        return [
            TestAdapter.from_protobuf(test_adapter) for test_adapter in query_response.test_adapters
        ]

    def register_schema_from_file(self, schema_file_path: Path | str) -> str:
        """Register a schema obtained from the specified file in the metadata store.

        Args:
            schema_file_path: The path at which the schema file is located

        Raises:
            FileNotFoundError: If the schema file does not exist.
        """
        if isinstance(schema_file_path, str):
            schema_file_path = Path(schema_file_path)

        if not schema_file_path.exists():
            raise FileNotFoundError(f"Schema file not found: {schema_file_path}")

        schema_contents = schema_file_path.read_text(encoding="utf-8-sig")
        return self.register_schema(schema_contents=schema_contents)

    def register_schema(self, schema_contents: str) -> str:
        """Register a schema in the metadata store.

        Once a schema has been published, it cannot be modified or removed.

        Args:
            schema_contents: The contents of the JSON or TOML schema. This
                should be a well-formed JSON or TOML schema. Validation will
                be performed, and an error will be returned if the schema is
                not valid.

        Returns:
            str: The ID of the schema.
        """
        register_request = RegisterSchemaRequest(schema=schema_contents)
        register_response = self._get_metadata_store_client().register_schema(register_request)
        return register_response.schema_id

    def list_schemas(self) -> Sequence[ExtensionSchema]:
        """List the schemas that have been previously registered.

        Returns:
            Sequence[ExtensionSchema]: The list of registered schemas.
        """
        list_request = ListSchemasRequest()
        list_response = self._get_metadata_store_client().list_schemas(list_request)
        return [ExtensionSchema.from_protobuf(schema) for schema in list_response.schemas]

    def get_alias(self, alias_name: str) -> Alias:
        """Get an alias and its target (the underlying metadata it represents).

        Args:
            alias_name: The name of the alias to retrieve.

        Returns:
            Alias: The alias containing the alias name, target type, and
                target ID of the underlying metadata.
        """
        get_request = GetAliasRequest(alias_name=alias_name)
        get_response = self._get_metadata_store_client().get_alias(get_request)
        return Alias.from_protobuf(get_response.alias)

    def query_aliases(self, odata_query: str = "") -> Sequence[Alias]:
        """Perform an OData query on the registered aliases.

        Args:
            odata_query: An OData query string. Example: "$filter=name eq
                'Value'". $expand is not supported.

        Returns:
            Sequence[Alias]: The list of aliases that match the query.
        """
        query_request = QueryAliasesRequest(odata_query=odata_query)
        query_response = self._get_metadata_store_client().query_aliases(query_request)
        return [Alias.from_protobuf(alias) for alias in query_response.aliases]

    def create_alias(
        self,
        alias_name: str,
        alias_target: (
            UutInstance
            | Uut
            | HardwareItem
            | SoftwareItem
            | Operator
            | TestDescription
            | Test
            | TestAdapter
            | TestStation
        ),
    ) -> Alias:
        """Create (register) an alias of the specified metadata.

        The specified metadata must have already been created prior to the
        alias registration. This method may be called with an already
        registered alias name in order to update the target mapped for that
        existing alias.

        Args:
            alias_name: The alias name to register.

            alias_target: The metadata instance to alias. The metadata
                instance to alias must have already been created in the
                metadata store in order to register an alias for it.

        Returns:
            Alias: The created alias containing the alias name, target type,
                and target ID.
        """
        create_request = CreateAliasRequest(alias_name=alias_name)
        if isinstance(alias_target, UutInstance):
            create_request.uut_instance.CopyFrom(alias_target.to_protobuf())
        elif isinstance(alias_target, Uut):
            create_request.uut.CopyFrom(alias_target.to_protobuf())
        elif isinstance(alias_target, HardwareItem):
            create_request.hardware_item.CopyFrom(alias_target.to_protobuf())
        elif isinstance(alias_target, SoftwareItem):
            create_request.software_item.CopyFrom(alias_target.to_protobuf())
        elif isinstance(alias_target, Operator):
            create_request.operator.CopyFrom(alias_target.to_protobuf())
        elif isinstance(alias_target, TestDescription):
            create_request.test_description.CopyFrom(alias_target.to_protobuf())
        elif isinstance(alias_target, Test):
            create_request.test.CopyFrom(alias_target.to_protobuf())
        elif isinstance(alias_target, TestAdapter):
            create_request.test_adapter.CopyFrom(alias_target.to_protobuf())
        elif isinstance(alias_target, TestStation):
            create_request.test_station.CopyFrom(alias_target.to_protobuf())
        response = self._get_metadata_store_client().create_alias(create_request)
        return Alias.from_protobuf(response.alias)

    def delete_alias(self, alias_name: str) -> bool:
        """Remove a registered alias.

        Args:
            alias_name: The name of the alias to unregister.

        Returns:
            bool: Whether the action resulted in the specified alias becoming
                unregistered. False if the alias does not exist.
        """
        delete_request = DeleteAliasRequest(alias_name=alias_name)
        delete_response = self._get_metadata_store_client().delete_alias(delete_request)
        return delete_response.unregistered

    def create_from_json_file(self, metadata_file_path: Path | str) -> MetadataItems:
        """Create metadata items from a JSON file.

        Args:
            metadata_file_path: The path to the JSON file containing metadata definitions.

        Returns:
            MetadataItems: A collection of metadata items created from the JSON document.

        Raises:
            FileNotFoundError: If the JSON file does not exist.
        """
        if isinstance(metadata_file_path, str):
            metadata_file_path = Path(metadata_file_path)

        if not metadata_file_path.exists():
            raise FileNotFoundError(f"Metadata file not found: {metadata_file_path}")

        metadata_contents = metadata_file_path.read_text(encoding="utf-8-sig")
        return self.create_from_json(metadata_contents)

    def create_from_json(self, metadata_file_contents: str) -> MetadataItems:
        """Create metadata items from a JSON document.

        Args:
            metadata_file_contents: The JSON document content containing metadata definitions.

        Returns:
            MetadataItems: A collection of metadata items created from the JSON document.
        """
        create_request = CreateFromJsonDocumentRequest(json_document=metadata_file_contents)
        create_response = self._get_metadata_store_client().create_from_json_document(
            create_request
        )
        return MetadataItems.from_protobuf(create_response)

    def _get_metadata_store_client(self) -> MetadataStoreServiceClient:
        if self._closed:
            raise RuntimeError(self._METADATA_STORE_CLIENT_CLOSED_ERROR)

        if self._metadata_store_client is None:
            with self._metadata_store_client_lock:
                if self._metadata_store_client is None:
                    self._metadata_store_client = self._instantiate_metadata_store_client()
        return self._metadata_store_client

    def _instantiate_metadata_store_client(self) -> MetadataStoreServiceClient:
        return MetadataStoreServiceClient(
            discovery_client=self._discovery_client,
            grpc_channel=self._grpc_channel,
            grpc_channel_pool=self._grpc_channel_pool,
        )



"""Data store client for publishing and reading data."""

from __future__ import annotations

import logging
import sys
from collections.abc import Iterable, Sequence
from threading import Lock
from types import TracebackType
from typing import TYPE_CHECKING, Type, TypeVar, overload
from urllib.parse import urlparse

import hightime as ht
from grpc import Channel
from ni.datamonikers.v1.client import MonikerClient
from ni.datastore.data._grpc_conversion import (
    get_publish_measurement_timestamp,
    populate_publish_condition_batch_request_values,
    populate_publish_condition_request_value,
    populate_publish_measurement_batch_request_values,
    populate_publish_measurement_request_value,
    unpack_and_convert_from_protobuf_any,
)
from ni.datastore.data._types._error_information import ErrorInformation
from ni.datastore.data._types._moniker import Moniker
from ni.datastore.data._types._outcome import Outcome
from ni.datastore.data._types._published_condition import PublishedCondition
from ni.datastore.data._types._published_measurement import PublishedMeasurement
from ni.datastore.data._types._step import Step
from ni.datastore.data._types._test_result import TestResult
from ni.measurementlink.discovery.v1.client import DiscoveryClient
from ni.measurements.data.v1.client import DataStoreClient as DataStoreServiceClient
from ni.measurements.data.v1.data_store_service_pb2 import (
    CreateStepRequest,
    CreateTestResultRequest,
    GetConditionRequest,
    GetMeasurementRequest,
    GetStepRequest,
    GetTestResultRequest,
    PublishConditionBatchRequest,
    PublishConditionRequest,
    PublishMeasurementBatchRequest,
    PublishMeasurementRequest,
    QueryConditionsRequest,
    QueryMeasurementsRequest,
    QueryStepsRequest,
    QueryTestResultsRequest,
)
from ni.protobuf.types.precision_timestamp_conversion import (
    hightime_datetime_to_protobuf,
)
from ni_grpc_extensions.channelpool import GrpcChannelPool

if TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

TRead = TypeVar("TRead")

_logger = logging.getLogger(__name__)


class DataStoreClient:
    """Data store client for publishing and reading data."""

    __slots__ = (
        "_closed",
        "_discovery_client",
        "_grpc_channel",
        "_grpc_channel_pool",
        "_data_store_client",
        "_data_store_client_lock",
        "_moniker_clients_by_service_location",
        "_moniker_clients_lock",
    )

    _DATA_STORE_CLIENT_CLOSED_ERROR = (
        "This DataStoreClient has been closed. Create a new DataStoreClient for further "
        "interaction with the data store."
    )

    _closed: bool
    _discovery_client: DiscoveryClient | None
    _grpc_channel: Channel | None
    _grpc_channel_pool: GrpcChannelPool | None
    _data_store_client: DataStoreServiceClient | None
    _moniker_clients_by_service_location: dict[str, MonikerClient]
    _data_store_client_lock: Lock
    _moniker_clients_lock: Lock

    def __init__(
        self,
        discovery_client: DiscoveryClient | None = None,
        grpc_channel: Channel | None = None,
        grpc_channel_pool: GrpcChannelPool | None = None,
    ) -> None:
        """Initialize the DataStoreClient.

        Args:
            discovery_client: An optional discovery client (recommended).

            grpc_channel: An optional data store gRPC channel. Providing this channel will bypass
                discovery service resolution of the data store. (Note: Reading data from a moniker
                will still always use a channel corresponding to the service location specified by
                that moniker.)

            grpc_channel_pool: An optional gRPC channel pool (recommended).
        """
        self._discovery_client = discovery_client
        self._grpc_channel = grpc_channel
        self._grpc_channel_pool = grpc_channel_pool

        self._data_store_client = None
        self._moniker_clients_by_service_location = {}

        self._data_store_client_lock = Lock()
        self._moniker_clients_lock = Lock()

        self._closed = False

    def __enter__(self) -> Self:
        """Enter the runtime context of the data store client."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit the runtime context of the data store client."""
        self.close()

    def close(self) -> None:
        """Close the data store client and clean up resources that it owns."""
        self._closed = True

        with self._data_store_client_lock:
            if self._data_store_client is not None:
                self._data_store_client.close()
                self._data_store_client = None

        with self._moniker_clients_lock:
            for _, moniker_client in self._moniker_clients_by_service_location.items():
                moniker_client.close()
            self._moniker_clients_by_service_location.clear()

    def publish_condition(
        self,
        name: str,
        condition_type: str,
        value: object,
        step_id: str,
    ) -> str:
        """Publish a condition value to the data store.

        Args:
            name: An identifier describing the condition value.
                For example, "Voltage" or "Temperature".

            condition_type: The type of this condition. For example, "Upper Limit",
                "Environment", or "Setup".

            value: The single value for this condition to publish on the test
                step. This should be a scalar value that can be converted to
                the appropriate protobuf scalar type.

            step_id: The ID of the step associated with this condition. This
                value is expected to be a parsable GUID.

        Returns:
            str: The condition id - the unique ID of the condition for
                referencing in queries
        """
        publish_request = PublishConditionRequest(
            name=name,
            condition_type=condition_type,
            step_id=step_id,
        )
        populate_publish_condition_request_value(publish_request, value)
        publish_response = self._get_data_store_client().publish_condition(publish_request)
        return publish_response.condition_id

    def publish_condition_batch(
        self, name: str, condition_type: str, values: object, step_id: str
    ) -> str:
        """Publish a batch of N values for a condition to the data store.

        Args:
            name: An identifier describing the condition values.
                For example, "Voltage" or "Temperature".

            condition_type: The type of this condition. For example, "Upper Limit",
                "Environment", or "Setup".

            values: The values for this condition across all publishes on the
                test step. This should be a Vector of N values.

            step_id: The ID of the step associated with this batch of condition
                values. This value is expected to be a parsable GUID.

        Returns:
            str: The condition id - the unique ID of the condition for
                referencing in queries
        """
        publish_request = PublishConditionBatchRequest(
            name=name,
            condition_type=condition_type,
            step_id=step_id,
        )
        populate_publish_condition_batch_request_values(publish_request, values)
        publish_response = self._get_data_store_client().publish_condition_batch(publish_request)
        return publish_response.condition_id

    def publish_measurement(
        self,
        name: str,
        value: object,  # More strongly typed Union[bool, AnalogWaveform] can be used if needed
        step_id: str,
        timestamp: ht.datetime | None = None,
        outcome: Outcome = Outcome.UNSPECIFIED,
        error_information: ErrorInformation | None = None,
        hardware_item_ids: Iterable[str] = tuple(),
        test_adapter_ids: Iterable[str] = tuple(),
        software_item_ids: Iterable[str] = tuple(),
        notes: str = "",
    ) -> str:
        """Publish a single measurement value associated with a test step.

        Args:
            name: The name used for associating/grouping
                conceptually alike measurements across multiple publish
                iterations. For example, "Temperature" can be used for
                associating temperature readings across multiple iterations.

            value: The value of the measurement being published. Supported types:

                - Scalar: Single float, int, str or boolean
                - Vector: Array of float, int, str or boolean values
                - DoubleAnalogWaveform: Analog waveform with double precision
                - DoubleXYData: XY coordinate data with double precision
                - I16AnalogWaveform: Analog waveform with 16-bit integer precision
                - DoubleComplexWaveform: Complex waveform with double precision
                - I16ComplexWaveform: Complex waveform with 16-bit integer precision
                - DoubleSpectrum: Frequency spectrum data with double precision
                - DigitalWaveform: Digital waveform data

            step_id: The ID of the step associated with this measurement. This
                value is expected to be a parsable GUID.

            timestamp: The timestamp of the measurement. If None, the current
                time will be used.

            outcome: The outcome of the measurement (PASSED, FAILED,
                INDETERMINATE, or UNSPECIFIED).

            error_information: Error or exception information in case of
                measurement failure.

            hardware_item_ids: The IDs of the hardware items associated with
                this measurement. These values are expected to be parsable
                GUIDs or aliases.

            test_adapter_ids: The IDs of the test adapters associated with this
                measurement. These values are expected to be parsable GUIDs or
                aliases.

            software_item_ids: The IDs of the software items associated with
                this measurement. These values are expected to be parsable
                GUIDs or aliases.

            notes: Any notes to be associated with the captured measurement.

        Returns:
            str: The published measurement id.
        """
        publish_request = PublishMeasurementRequest(
            name=name,
            step_id=step_id,
            outcome=outcome.to_protobuf(),
            error_information=(
                error_information.to_protobuf() if error_information is not None else None
            ),
            hardware_item_ids=hardware_item_ids,
            test_adapter_ids=test_adapter_ids,
            software_item_ids=software_item_ids,
            notes=notes,
        )
        populate_publish_measurement_request_value(publish_request, value)
        publish_request.timestamp.CopyFrom(
            get_publish_measurement_timestamp(publish_request, timestamp)
        )
        publish_response = self._get_data_store_client().publish_measurement(publish_request)
        return publish_response.measurement_id

    def publish_measurement_batch(
        self,
        name: str,
        values: object,
        step_id: str,
        timestamps: Iterable[ht.datetime] = tuple(),
        outcomes: Iterable[Outcome] = tuple(),
        error_information: Iterable[ErrorInformation] = tuple(),
        hardware_item_ids: Iterable[str] = tuple(),
        test_adapter_ids: Iterable[str] = tuple(),
        software_item_ids: Iterable[str] = tuple(),
        notes: str = "",
    ) -> Sequence[str]:
        """Publish multiple scalar measurements at once for parametric sweeps.

        Args:
            name: The name used for associating/grouping
                conceptually alike measurements across multiple publish
                iterations. For example, "Temperature" can be used for
                associating temperature readings across multiple iterations.

            values: The values of the (scalar) measurement being published
                across N iterations.

            step_id: The ID of the step associated with this measurement. This
                value is expected to be a parsable GUID.

            timestamps: The timestamps corresponding to the N iterations of
                batched measurement being published. Can be empty (no timestamp
                info), single value (applied to all), or N values (one per
                measurement).

            outcomes: The outcomes corresponding to the N iterations of batched
                measurement being published. Can be empty (no outcome info),
                single value (applied to all), or N values (one per
                measurement).

            error_information: The error information corresponding to the N
                iterations of batched measurement being published. Can be empty
                (no error info), single value (applied to all), or N values
                (one per measurement).

            hardware_item_ids: The IDs of the hardware items associated with
                this measurement. These values are expected to be parsable
                GUIDs or aliases.

            test_adapter_ids: The IDs of the test adapters associated with this
                measurement. These values are expected to be parsable GUIDs or
                aliases.

            software_item_ids: The IDs of the software items associated with
                this measurement. These values are expected to be parsable
                GUIDs or aliases.

            notes: Any notes to be associated with the published measurements.

        Returns:
            Sequence[str]: The ids of the published measurement ids.
                NOTE: Using a Sequence is for future flexibility.
                This sequence will currently always have a single measurement id
                returned.
        """
        publish_request = PublishMeasurementBatchRequest(
            name=name,
            step_id=step_id,
            timestamps=[hightime_datetime_to_protobuf(ts) for ts in timestamps],
            outcomes=[outcome.to_protobuf() for outcome in outcomes],
            error_information=(
                [ei.to_protobuf() for ei in (error_information or [])] if error_information else []
            ),
            hardware_item_ids=hardware_item_ids,
            test_adapter_ids=test_adapter_ids,
            software_item_ids=software_item_ids,
            notes=notes,
        )
        populate_publish_measurement_batch_request_values(publish_request, values)
        publish_response = self._get_data_store_client().publish_measurement_batch(publish_request)
        return publish_response.measurement_ids

    @overload
    def read_data(
        self,
        moniker_source: Moniker | PublishedMeasurement | PublishedCondition,
        expected_type: Type[TRead],
    ) -> TRead: ...

    @overload
    def read_data(
        self,
        moniker_source: Moniker | PublishedMeasurement | PublishedCondition,
    ) -> object: ...

    def read_data(
        self,
        moniker_source: Moniker | PublishedMeasurement | PublishedCondition,
        expected_type: Type[TRead] | None = None,
    ) -> TRead | object:
        """Read data published to the data store.

        Args:
            moniker_source: The source from which to read data. Can be:
                - A Moniker (wrapper type) directly
                - A PublishedMeasurement (uses its moniker)
                - A PublishedCondition (uses its moniker)

            expected_type: Optional type to validate the returned data against.
                If provided, a TypeError will be raised if the actual data type
                doesn't match.

        Returns:
            The data retrieved from the data store. The return type depends on
            what was originally published:
            - Scalar measurements return as Vectors
            - Other types are returned as originally published
            If expected_type is specified, the return value is guaranteed to be
            of that type.

        Raises:
            ValueError: If the moniker_source doesn't have a valid moniker.
            TypeError: If expected_type is provided and the actual data type
                doesn't match.
        """
        from ni.datamonikers.v1.data_moniker_pb2 import Moniker as MonikerProto

        moniker_proto: MonikerProto

        if isinstance(moniker_source, Moniker):
            moniker_proto = moniker_source.to_protobuf()
        elif isinstance(moniker_source, PublishedMeasurement):
            if moniker_source.moniker is None:
                raise ValueError("PublishedMeasurement must have a Moniker to read data")
            moniker_proto = moniker_source.moniker.to_protobuf()
        elif isinstance(moniker_source, PublishedCondition):
            if moniker_source.moniker is None:
                raise ValueError("PublishedCondition must have a Moniker to read data")
            moniker_proto = moniker_source.moniker.to_protobuf()
        else:
            raise TypeError(f"Unsupported moniker_source type: {type(moniker_source)}")

        moniker_client = self._get_moniker_client(moniker_proto.service_location)
        read_result = moniker_client.read_from_moniker(moniker_proto)
        converted_data = unpack_and_convert_from_protobuf_any(read_result.value)
        if expected_type is not None and not isinstance(converted_data, expected_type):
            raise TypeError(f"Expected type {expected_type}, got {type(converted_data)}")
        return converted_data

    def create_step(self, step: Step) -> str:
        """Create a new step in the data store.

        A step is owned by a test result and is a logical grouping of published
        measurements and conditions. All measurements and conditions must be
        associated with a step.

        Args:
            step: The metadata of the step to be created.

        Returns:
            str: The identifier of the created step.
        """
        create_request = CreateStepRequest(step=step.to_protobuf())
        create_response = self._get_data_store_client().create_step(create_request)
        return create_response.step_id

    def get_step(self, step_id: str) -> Step:
        """Get the step associated with the given identifier.

        Args:
            step_id: The identifier of the desired step.

        Returns:
            Step: The metadata of the requested step.
        """
        get_request = GetStepRequest(step_id=step_id)
        get_response = self._get_data_store_client().get_step(get_request)
        return Step.from_protobuf(get_response.step)

    def get_measurement(self, measurement_id: str) -> PublishedMeasurement:
        """Get the measurement associated with the given identifier.

        Args:
            measurement_id: The identifier of the desired measurement.

        Returns:
            PublishedMeasurement: The metadata of the requested measurement.
        """
        get_request = GetMeasurementRequest(measurement_id=measurement_id)
        get_response = self._get_data_store_client().get_measurement(get_request)
        return PublishedMeasurement.from_protobuf(get_response.published_measurement)

    def get_condition(self, condition_id: str) -> PublishedCondition:
        """Get the condition associated with the given identifier.

        Args:
            condition_id: The identifier of the desired condition.

        Returns:
            PublishedCondition: The metadata of the requested condition.
        """
        get_request = GetConditionRequest(condition_id=condition_id)
        get_response = self._get_data_store_client().get_condition(get_request)
        return PublishedCondition.from_protobuf(get_response.published_condition)

    def create_test_result(self, test_result: TestResult) -> str:
        """Create a test result object for publishing measurements.

        Once a test result is created, you can publish an arbitrary number of
        measurements and conditions to a step which is owned by the test result.

        Args:
            test_result: The metadata of the test result to be created.

        Returns:
            str: The test result ID. Generated if not specified in the request.
        """
        create_request = CreateTestResultRequest(test_result=test_result.to_protobuf())
        create_response = self._get_data_store_client().create_test_result(create_request)
        return create_response.test_result_id

    def get_test_result(self, test_result_id: str) -> TestResult:
        """Get the test result associated with the given identifier.

        Args:
            test_result_id: The ID of the desired test result. This value is
                expected to be a parsable GUID.

        Returns:
            TestResult: The TestResult object that corresponds to the
                requested ID.
        """
        get_request = GetTestResultRequest(test_result_id=test_result_id)
        get_response = self._get_data_store_client().get_test_result(get_request)
        return TestResult.from_protobuf(get_response.test_result)

    def query_conditions(self, odata_query: str = "") -> Sequence[PublishedCondition]:
        """Query conditions using OData query syntax.

        Args:
            odata_query: An OData query string. Example: "$filter=name eq
                'Value'". An empty string will return all conditions. $expand,
                $count, and $select are not supported. For more information,
                see https://learn.microsoft.com/en-us/odata/concepts/
                queryoptions-overview.

        Returns:
            Sequence[PublishedCondition]: The list of matching conditions. Each
                item contains a moniker for retrieving the condition
                measurements, as well as the metadata associated with the
                condition.
        """
        query_request = QueryConditionsRequest(odata_query=odata_query)
        query_response = self._get_data_store_client().query_conditions(query_request)
        return [
            PublishedCondition.from_protobuf(published_condition)
            for published_condition in query_response.published_conditions
        ]

    def query_measurements(self, odata_query: str = "") -> Sequence[PublishedMeasurement]:
        """Query measurements using OData query syntax.

        Args:
            odata_query: An OData query string. Example: "$filter=name eq
                'Value'". An empty string will return all measurements.
                $expand, $count, and $select are not supported. For more
                information, see https://learn.microsoft.com/en-us/odata/
                concepts/queryoptions-overview.

        Returns:
            Sequence[PublishedMeasurement]: The list of matching measurements.
                Each item contains a moniker for retrieving the measurement, as
                well as the metadata associated with the measurement.
        """
        query_request = QueryMeasurementsRequest(odata_query=odata_query)
        query_response = self._get_data_store_client().query_measurements(query_request)
        return [
            PublishedMeasurement.from_protobuf(published_measurement)
            for published_measurement in query_response.published_measurements
        ]

    def query_test_results(self, odata_query: str = "") -> Sequence[TestResult]:
        """Query test results using OData query syntax.

        Args:
            odata_query: An OData query string. Example: "$filter=name eq
                'Value'". An empty string will return all test results.
                $expand, $count, and $select are not supported. For more
                information, see https://learn.microsoft.com/en-us/odata/
                concepts/queryoptions-overview.

        Returns:
            Sequence[TestResult]: The list of matching test results. Each
                item contains the metadata associated with the test result,
                including test result ID, name, timestamps, and other
                properties.
        """
        query_request = QueryTestResultsRequest(odata_query=odata_query)
        query_response = self._get_data_store_client().query_test_results(query_request)
        return [
            TestResult.from_protobuf(test_result) for test_result in query_response.test_results
        ]

    def query_steps(self, odata_query: str = "") -> Sequence[Step]:
        """Query for steps matching the given OData query.

        Args:
            odata_query: An OData query string. Example: "$filter=name eq
                'Value'". An empty string will return all steps. $expand,
                $count, and $select are not supported. For more information,
                see https://learn.microsoft.com/en-us/odata/concepts/
                queryoptions-overview.

        Returns:
            Sequence[Step]: The list of steps that match the query.
        """
        query_request = QueryStepsRequest(odata_query=odata_query)
        query_response = self._get_data_store_client().query_steps(query_request)
        return [Step.from_protobuf(step) for step in query_response.steps]

    def _get_data_store_client(self) -> DataStoreServiceClient:
        if self._closed:
            raise RuntimeError(self._DATA_STORE_CLIENT_CLOSED_ERROR)

        if self._data_store_client is None:
            with self._data_store_client_lock:
                if self._data_store_client is None:
                    self._data_store_client = self._instantiate_data_store_client()
        return self._data_store_client

    def _instantiate_data_store_client(self) -> DataStoreServiceClient:
        return DataStoreServiceClient(
            discovery_client=self._discovery_client,
            grpc_channel=self._grpc_channel,
            grpc_channel_pool=self._grpc_channel_pool,
        )

    def _get_moniker_client(self, service_location: str) -> MonikerClient:
        if self._closed:
            raise RuntimeError(self._DATA_STORE_CLIENT_CLOSED_ERROR)

        parsed_service_location = urlparse(service_location).netloc
        if parsed_service_location not in self._moniker_clients_by_service_location:
            with self._moniker_clients_lock:
                if parsed_service_location not in self._moniker_clients_by_service_location:
                    self._moniker_clients_by_service_location[parsed_service_location] = (
                        self._instantiate_moniker_client(parsed_service_location)
                    )
        return self._moniker_clients_by_service_location[parsed_service_location]

    def _instantiate_moniker_client(self, parsed_service_location: str) -> MonikerClient:
        return MonikerClient(
            service_location=parsed_service_location,
            grpc_channel_pool=self._grpc_channel_pool,
        )



"""Test data type for the Data Store Client."""

from __future__ import annotations

from typing import Mapping, MutableMapping

from ni.datastore.metadata._grpc_conversion import (
    populate_extension_value_message_map,
    populate_from_extension_value_message_map,
)
from ni.measurements.metadata.v1.metadata_store_pb2 import (
    Test as TestProto,
)


class Test:
    """Represents the metadata of a test.

    A test contains information about a specific test procedure, including its
    name and description of what the test does.
    """

    __slots__ = (
        "_id",
        "name",
        "description",
        "link",
        "_extension",
        "schema_id",
    )

    @property
    def extension(self) -> MutableMapping[str, str]:
        """The extension of the test."""
        return self._extension

    @property
    def id(self) -> str:
        """The string id of the test."""
        return self._id

    def __init__(
        self,
        *,
        name: str = "",
        description: str = "",
        link: str = "",
        extension: Mapping[str, str] | None = None,
        schema_id: str = "",
    ) -> None:
        """Initialize a Test instance.

        Args:
            name: The name of the test.
            description: A description of what the test does.
            link: A link to a resource that describes the test. This
                value is expected to be a valid URI.
            extension: Any extensions to be associated with the test.
            schema_id: The unique identifier of the schema that applies to this
                instance's extension. If any extension is associated with this
                instance, a schema_id must be provided, unless the test
                is created within the context of a test result, in which case
                the test result must have a schema_id.
        """
        self._id = ""
        self.name = name
        self.description = description
        self.link = link
        self._extension: MutableMapping[str, str] = dict(extension) if extension is not None else {}
        self.schema_id = schema_id

    @staticmethod
    def from_protobuf(test_proto: TestProto) -> "Test":
        """Create a Test instance from a protobuf Test message."""
        test = Test(
            name=test_proto.name,
            description=test_proto.description,
            link=test_proto.link,
            schema_id=test_proto.schema_id,
        )
        populate_from_extension_value_message_map(test.extension, test_proto.extension)
        test._id = test_proto.id
        return test

    def to_protobuf(self) -> TestProto:
        """Convert this Test to a protobuf Test message."""
        test_proto = TestProto(
            id=self.id,
            name=self.name,
            description=self.description,
            link=self.link,
            schema_id=self.schema_id,
        )
        populate_extension_value_message_map(test_proto.extension, self.extension)
        return test_proto

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, Test):
            return NotImplemented
        return (
            self.id == other.id
            and self.name == other.name
            and self.description == other.description
            and self.link == other.link
            and self.extension == other.extension
            and self.schema_id == other.schema_id
        )

    def __str__(self) -> str:
        """Return a string representation of the Test."""
        return str(self.to_protobuf())



"""Test Adapter data type for the Data Store Client."""

from __future__ import annotations

from typing import Mapping, MutableMapping

from ni.datastore.metadata._grpc_conversion import (
    populate_extension_value_message_map,
    populate_from_extension_value_message_map,
)
from ni.measurements.metadata.v1.metadata_store_pb2 import (
    TestAdapter as TestAdapterProto,
)


class TestAdapter:
    """Represents a test adapter or mechanical setup.

    This is a board or device that is used to hold, connect, or interface the
    UUT with the test system. Test adapters may also be referred to as test
    fixtures, interface boards, breakout boxes, mechanical fixtures, or
    connection adapters.
    """

    __slots__ = (
        "_id",
        "name",
        "manufacturer",
        "model",
        "serial_number",
        "part_number",
        "asset_identifier",
        "calibration_due_date",
        "link",
        "_extension",
        "schema_id",
    )

    @property
    def extension(self) -> MutableMapping[str, str]:
        """The extension of the test adapter."""
        return self._extension

    @property
    def id(self) -> str:
        """The string id of the test adapter."""
        return self._id

    def __init__(
        self,
        *,
        name: str = "",
        manufacturer: str = "",
        model: str = "",
        serial_number: str = "",
        part_number: str = "",
        asset_identifier: str = "",
        calibration_due_date: str = "",
        link: str = "",
        extension: Mapping[str, str] | None = None,
        schema_id: str = "",
    ) -> None:
        """Initialize a TestAdapter instance.

        Args:
            name: The name of the test adapter.
            manufacturer: The manufacturer of the test adapter.
            model: The model of the test adapter.
            serial_number: The serial number of the test adapter.
            part_number: The part number of the test adapter.
            asset_identifier: The asset identifier of the test adapter.
            calibration_due_date: The calibration due date of the test adapter.
            link: A link to a resource that describes the test adapter. This
                value is expected to be a valid URI.
            extension: Any extensions to be associated with the test adapter.
            schema_id: The unique identifier of the schema that applies to this
                instance's extension. If any extension is associated with this
                instance, a schema_id must be provided, unless the test adapter
                is created within the context of a test result, in which case
                the test result must have a schema_id.
        """
        self._id = ""
        self.name = name
        self.manufacturer = manufacturer
        self.model = model
        self.serial_number = serial_number
        self.part_number = part_number
        self.asset_identifier = asset_identifier
        self.calibration_due_date = calibration_due_date
        self.link = link
        self._extension: MutableMapping[str, str] = dict(extension) if extension is not None else {}
        self.schema_id = schema_id

    @staticmethod
    def from_protobuf(test_adapter_proto: TestAdapterProto) -> "TestAdapter":
        """Create a TestAdapter instance from a protobuf TestAdapter message."""
        test_adapter = TestAdapter(
            name=test_adapter_proto.name,
            manufacturer=test_adapter_proto.manufacturer,
            model=test_adapter_proto.model,
            serial_number=test_adapter_proto.serial_number,
            part_number=test_adapter_proto.part_number,
            asset_identifier=test_adapter_proto.asset_identifier,
            calibration_due_date=test_adapter_proto.calibration_due_date,
            link=test_adapter_proto.link,
            schema_id=test_adapter_proto.schema_id,
        )
        populate_from_extension_value_message_map(
            test_adapter.extension, test_adapter_proto.extension
        )
        test_adapter._id = test_adapter_proto.id
        return test_adapter

    def to_protobuf(self) -> TestAdapterProto:
        """Convert this TestAdapter to a protobuf TestAdapter message."""
        test_adapter_proto = TestAdapterProto(
            id=self.id,
            name=self.name,
            manufacturer=self.manufacturer,
            model=self.model,
            serial_number=self.serial_number,
            part_number=self.part_number,
            asset_identifier=self.asset_identifier,
            calibration_due_date=self.calibration_due_date,
            link=self.link,
            schema_id=self.schema_id,
        )
        populate_extension_value_message_map(test_adapter_proto.extension, self.extension)
        return test_adapter_proto

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, TestAdapter):
            return NotImplemented
        return (
            self.id == other.id
            and self.name == other.name
            and self.manufacturer == other.manufacturer
            and self.model == other.model
            and self.serial_number == other.serial_number
            and self.part_number == other.part_number
            and self.asset_identifier == other.asset_identifier
            and self.calibration_due_date == other.calibration_due_date
            and self.link == other.link
            and self.extension == other.extension
            and self.schema_id == other.schema_id
        )

    def __str__(self) -> str:
        """Return a string representation of the TestAdapter."""
        return str(self.to_protobuf())



"""Published Measurement data type for the Data Store Client."""

from __future__ import annotations

from typing import Iterable, MutableSequence

import hightime as ht
from ni.datastore.data._types._error_information import ErrorInformation
from ni.datastore.data._types._moniker import Moniker
from ni.datastore.data._types._outcome import Outcome
from ni.datastore.data._types._published_condition import PublishedCondition
from ni.measurements.data.v1.data_store_pb2 import PublishedMeasurement as PublishedMeasurementProto
from ni.protobuf.types.precision_timestamp_conversion import (
    hightime_datetime_from_protobuf,
    hightime_datetime_to_protobuf,
)


class PublishedMeasurement:
    """Represents a measurement that has been published to the data store.

    A published measurement contains the measurement data and the associated
    metadata (name, type, timestamps, outcome, hardware, software, and test
    adapter IDs).
    """

    __slots__ = (
        "moniker",
        "_published_conditions",
        "id",
        "test_result_id",
        "step_id",
        "_software_item_ids",
        "_hardware_item_ids",
        "_test_adapter_ids",
        "name",
        "value_type",
        "notes",
        "start_date_time",
        "end_date_time",
        "outcome",
        "parametric_index",
        "error_information",
    )

    @property
    def published_conditions(self) -> MutableSequence[PublishedCondition]:
        """The published conditions associated with the published measurement."""
        return self._published_conditions

    @property
    def software_item_ids(self) -> MutableSequence[str]:
        """The software item IDs associated with the published measurement."""
        return self._software_item_ids

    @property
    def hardware_item_ids(self) -> MutableSequence[str]:
        """The hardware item IDs associated with the published measurement."""
        return self._hardware_item_ids

    @property
    def test_adapter_ids(self) -> MutableSequence[str]:
        """The test adapter IDs associated with the published measurement."""
        return self._test_adapter_ids

    def __init__(
        self,
        *,
        moniker: Moniker | None = None,
        published_conditions: Iterable[PublishedCondition] | None = None,
        id: str = "",
        test_result_id: str = "",
        step_id: str = "",
        software_item_ids: Iterable[str] | None = None,
        hardware_item_ids: Iterable[str] | None = None,
        test_adapter_ids: Iterable[str] | None = None,
        name: str = "",
        value_type: str = "",
        notes: str = "",
        start_date_time: ht.datetime | None = None,
        end_date_time: ht.datetime | None = None,
        outcome: Outcome = Outcome.UNSPECIFIED,
        parametric_index: int = 0,
        error_information: ErrorInformation | None = None,
    ) -> None:
        """Initialize a PublishedMeasurement instance.

        Args:
            moniker: The moniker of the measurement that this value is
                associated with.
            published_conditions: The published conditions associated with this
                measurement from the test step.
            id: The unique identifier of the measurement.
                This can be used to reference and find the measurement in the
                data store.
            test_result_id: The ID of the test result with which this
                measurement is associated.
            step_id: The ID of the step with which this measurement is
                associated.
            software_item_ids: The IDs of the software items associated with
                this measurement.
            hardware_item_ids: The IDs of the hardware items associated with
                this measurement.
            test_adapter_ids: The IDs of the test adapters associated with this
                measurement.
            name: The name of the measurement.
            value_type: The data type of the measurement value.
            notes: Additional notes or comments about the
                measurement.
            start_date_time: The start timestamp of the measurement.
            end_date_time: The end timestamp of the measurement.
            outcome: The outcome of the measurement (PASSED, FAILED,
                INDETERMINATE, or UNSPECIFIED).
            parametric_index: The index of this measurement in a parametric
                sweep operation.
            error_information: Error or exception information in case of
                measurement failure.
        """
        self.moniker = moniker
        self._published_conditions: MutableSequence[PublishedCondition] = (
            list(published_conditions) if published_conditions is not None else []
        )
        self.id = id
        self.test_result_id = test_result_id
        self.step_id = step_id
        self._software_item_ids: MutableSequence[str] = (
            list(software_item_ids) if software_item_ids is not None else []
        )
        self._hardware_item_ids: MutableSequence[str] = (
            list(hardware_item_ids) if hardware_item_ids is not None else []
        )
        self._test_adapter_ids: MutableSequence[str] = (
            list(test_adapter_ids) if test_adapter_ids is not None else []
        )
        self.name = name
        self.value_type = value_type
        self.notes = notes
        self.start_date_time = start_date_time
        self.end_date_time = end_date_time
        self.outcome = outcome
        self.parametric_index = parametric_index
        self.error_information = error_information

    @staticmethod
    def from_protobuf(
        published_measurement_proto: PublishedMeasurementProto,
    ) -> "PublishedMeasurement":
        """Create a PublishedMeasurement instance from a protobuf PublishedMeasurement message."""
        return PublishedMeasurement(
            moniker=(
                Moniker.from_protobuf(published_measurement_proto.moniker)
                if published_measurement_proto.HasField("moniker")
                else None
            ),
            published_conditions=[
                PublishedCondition.from_protobuf(cond)
                for cond in published_measurement_proto.published_conditions
            ],
            id=published_measurement_proto.id,
            test_result_id=published_measurement_proto.test_result_id,
            step_id=published_measurement_proto.step_id,
            software_item_ids=published_measurement_proto.software_item_ids,
            hardware_item_ids=published_measurement_proto.hardware_item_ids,
            test_adapter_ids=published_measurement_proto.test_adapter_ids,
            name=published_measurement_proto.name,
            value_type=published_measurement_proto.value_type,
            notes=published_measurement_proto.notes,
            start_date_time=(
                hightime_datetime_from_protobuf(published_measurement_proto.start_date_time)
                if published_measurement_proto.HasField("start_date_time")
                else None
            ),
            end_date_time=(
                hightime_datetime_from_protobuf(published_measurement_proto.end_date_time)
                if published_measurement_proto.HasField("end_date_time")
                else None
            ),
            outcome=Outcome.from_protobuf(published_measurement_proto.outcome),
            parametric_index=published_measurement_proto.parametric_index,
            error_information=(
                ErrorInformation.from_protobuf(published_measurement_proto.error_information)
                if published_measurement_proto.HasField("error_information")
                else None
            ),
        )

    def to_protobuf(self) -> PublishedMeasurementProto:
        """Convert this PublishedMeasurement instance to a protobuf PublishedMeasurement message."""
        return PublishedMeasurementProto(
            moniker=self.moniker.to_protobuf() if self.moniker is not None else None,
            published_conditions=[
                condition.to_protobuf() for condition in self.published_conditions
            ],
            id=self.id,
            test_result_id=self.test_result_id,
            step_id=self.step_id,
            software_item_ids=self.software_item_ids,
            hardware_item_ids=self.hardware_item_ids,
            test_adapter_ids=self.test_adapter_ids,
            name=self.name,
            value_type=self.value_type,
            notes=self.notes,
            start_date_time=(
                hightime_datetime_to_protobuf(self.start_date_time)
                if self.start_date_time is not None
                else None
            ),
            end_date_time=(
                hightime_datetime_to_protobuf(self.end_date_time)
                if self.end_date_time is not None
                else None
            ),
            outcome=self.outcome.to_protobuf(),
            parametric_index=self.parametric_index,
            error_information=(
                self.error_information.to_protobuf() if self.error_information is not None else None
            ),
        )

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, PublishedMeasurement):
            return NotImplemented
        return (
            self.moniker == other.moniker
            and self.published_conditions == other.published_conditions
            and self.id == other.id
            and self.test_result_id == other.test_result_id
            and self.step_id == other.step_id
            and self.software_item_ids == other.software_item_ids
            and self.hardware_item_ids == other.hardware_item_ids
            and self.test_adapter_ids == other.test_adapter_ids
            and self.name == other.name
            and self.value_type == other.value_type
            and self.notes == other.notes
            and self.start_date_time == other.start_date_time
            and self.end_date_time == other.end_date_time
            and self.outcome == other.outcome
            and self.parametric_index == other.parametric_index
            and self.error_information == other.error_information
        )

    def __str__(self) -> str:
        """Return a string representation of the PublishedMeasurement."""
        return str(self.to_protobuf())



"""UUT Instance data type for the Data Store Client."""

from __future__ import annotations

from typing import Mapping, MutableMapping

from ni.datastore.metadata._grpc_conversion import (
    populate_extension_value_message_map,
    populate_from_extension_value_message_map,
)
from ni.measurements.metadata.v1.metadata_store_pb2 import (
    UutInstance as UutInstanceProto,
)


class UutInstance:
    """Represents the metadata of a UUT instance.

    A UUT instance represents a specific physical instance of a Unit Under
    Test, with properties like serial number, manufacture date, and firmware
    version that distinguish it from other instances of the same UUT model.


    """

    __slots__ = (
        "_id",
        "uut_id",
        "serial_number",
        "manufacture_date",
        "firmware_version",
        "hardware_version",
        "link",
        "_extension",
        "schema_id",
    )

    @property
    def extension(self) -> MutableMapping[str, str]:
        """The extension of the UUT instance."""
        return self._extension

    @property
    def id(self) -> str:
        """The string id of the uut instance."""
        return self._id

    def __init__(
        self,
        *,
        uut_id: str = "",
        serial_number: str = "",
        manufacture_date: str = "",
        firmware_version: str = "",
        hardware_version: str = "",
        link: str = "",
        extension: Mapping[str, str] | None = None,
        schema_id: str = "",
    ) -> None:
        """Initialize a UutInstance instance.

        Args:
            uut_id: The ID of the UUT associated with this UUT instance. This
                value is expected to be a parsable GUID or an alias.
            serial_number: The serial number of the UUT instance.
            manufacture_date: The date the UUT instance was manufactured.
            firmware_version: Version of the firmware on the UUT instance.
            hardware_version: Hardware version of the UUT instance.
            link: A link to a resource that describes the UUT instance. This
                value is expected to be a valid URI.
            extension: Any extensions to be associated with the UUT instance.
            schema_id: The unique identifier of the schema that applies to this
                instance's extension. If any extension is associated with this
                instance, a schema_id must be provided, unless the UUT instance
                is created within the context of a test result, in which case
                the test result must have a schema_id.
        """
        self._id = ""
        self.uut_id = uut_id
        self.serial_number = serial_number
        self.manufacture_date = manufacture_date
        self.firmware_version = firmware_version
        self.hardware_version = hardware_version
        self.link = link
        self._extension: MutableMapping[str, str] = dict(extension) if extension is not None else {}
        self.schema_id = schema_id

    @staticmethod
    def from_protobuf(uut_instance_proto: UutInstanceProto) -> "UutInstance":
        """Create a UutInstance from a protobuf UutInstance message."""
        uut_instance = UutInstance(
            uut_id=uut_instance_proto.uut_id,
            serial_number=uut_instance_proto.serial_number,
            manufacture_date=uut_instance_proto.manufacture_date,
            firmware_version=uut_instance_proto.firmware_version,
            hardware_version=uut_instance_proto.hardware_version,
            link=uut_instance_proto.link,
            schema_id=uut_instance_proto.schema_id,
        )
        populate_from_extension_value_message_map(
            uut_instance.extension, uut_instance_proto.extension
        )
        uut_instance._id = uut_instance_proto.id
        return uut_instance

    def to_protobuf(self) -> UutInstanceProto:
        """Convert this UutInstance to a protobuf UutInstance message."""
        uut_instance_proto = UutInstanceProto(
            id=self.id,
            uut_id=self.uut_id,
            serial_number=self.serial_number,
            manufacture_date=self.manufacture_date,
            firmware_version=self.firmware_version,
            hardware_version=self.hardware_version,
            link=self.link,
            schema_id=self.schema_id,
        )
        populate_extension_value_message_map(uut_instance_proto.extension, self.extension)
        return uut_instance_proto

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, UutInstance):
            return NotImplemented
        return (
            self.id == other.id
            and self.uut_id == other.uut_id
            and self.serial_number == other.serial_number
            and self.manufacture_date == other.manufacture_date
            and self.firmware_version == other.firmware_version
            and self.hardware_version == other.hardware_version
            and self.link == other.link
            and self.extension == other.extension
            and self.schema_id == other.schema_id
        )

    def __str__(self) -> str:
        """Return a string representation of the UutInstance."""
        return str(self.to_protobuf())



"""Extension Schema data type for the Data Store Client."""

from __future__ import annotations

from ni.measurements.metadata.v1.metadata_store_pb2 import (
    ExtensionSchema as ExtensionSchemaProto,
)


class ExtensionSchema:
    """Represents an extension schema stored in the system.

    An extension schema contains the schema ID and the schema definition
    itself, which can be used to validate extension data in metadata instances.
    """

    __slots__ = (
        "id",
        "schema",
    )

    def __init__(
        self,
        *,
        id: str = "",
        schema: str = "",
    ) -> None:
        """Initialize an ExtensionSchema instance.

        Args:
            id: The ID of the schema.
            schema: The schema itself.
        """
        self.id = id
        self.schema = schema

    @staticmethod
    def from_protobuf(extension_schema_proto: ExtensionSchemaProto) -> "ExtensionSchema":
        """Create an ExtensionSchema instance from a protobuf ExtensionSchema message."""
        return ExtensionSchema(
            id=extension_schema_proto.id,
            schema=extension_schema_proto.schema,
        )

    def to_protobuf(self) -> ExtensionSchemaProto:
        """Convert this ExtensionSchema to a protobuf ExtensionSchema message."""
        return ExtensionSchemaProto(
            id=self.id,
            schema=self.schema,
        )

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, ExtensionSchema):
            return NotImplemented
        return self.id == other.id and self.schema == other.schema

    def __str__(self) -> str:
        """Return a string representation of the ExtensionSchema."""
        return str(self.to_protobuf())



"""Helper methods to convert between python and protobuf types."""

from __future__ import annotations

import datetime as std_datetime
import logging
from typing import Iterable, cast

import hightime as ht
import numpy as np
from google.protobuf.any_pb2 import Any
from ni.measurements.data.v1.data_store_service_pb2 import (
    PublishConditionBatchRequest,
    PublishConditionRequest,
    PublishMeasurementBatchRequest,
    PublishMeasurementRequest,
)
from ni.protobuf.types.precision_timestamp_conversion import (
    hightime_datetime_to_protobuf,
)
from ni.protobuf.types.precision_timestamp_pb2 import PrecisionTimestamp
from ni.protobuf.types.scalar_conversion import scalar_to_protobuf
from ni.protobuf.types.vector_conversion import vector_from_protobuf, vector_to_protobuf
from ni.protobuf.types.vector_pb2 import Vector as VectorProto
from ni.protobuf.types.waveform_conversion import (
    digital_waveform_from_protobuf,
    digital_waveform_to_protobuf,
    float64_analog_waveform_from_protobuf,
    float64_analog_waveform_to_protobuf,
    float64_complex_waveform_from_protobuf,
    float64_complex_waveform_to_protobuf,
    float64_spectrum_from_protobuf,
    float64_spectrum_to_protobuf,
    int16_analog_waveform_from_protobuf,
    int16_analog_waveform_to_protobuf,
    int16_complex_waveform_from_protobuf,
    int16_complex_waveform_to_protobuf,
)
from ni.protobuf.types.waveform_pb2 import (
    DigitalWaveform as DigitalWaveformProto,
    DoubleAnalogWaveform,
    DoubleComplexWaveform,
    DoubleSpectrum,
    I16AnalogWaveform,
    I16ComplexWaveform,
)
from ni.protobuf.types.xydata_conversion import (
    float64_xydata_from_protobuf,
    float64_xydata_to_protobuf,
)
from ni.protobuf.types.xydata_pb2 import DoubleXYData
from nitypes.complex import ComplexInt32DType
from nitypes.scalar import Scalar
from nitypes.vector import Vector
from nitypes.waveform import AnalogWaveform, ComplexWaveform, DigitalWaveform, Spectrum
from nitypes.xy_data import XYData

_logger = logging.getLogger(__name__)


def populate_publish_condition_request_value(
    publish_request: PublishConditionRequest, value: object
) -> None:
    """Assign a value to the scalar member of PublishConditionRequest."""
    if isinstance(value, bool):
        publish_request.scalar.bool_value = value
    elif isinstance(value, int):
        publish_request.scalar.sint32_value = value
    elif isinstance(value, float):
        publish_request.scalar.double_value = value
    elif isinstance(value, str):
        publish_request.scalar.string_value = value
    elif isinstance(value, Scalar):
        publish_request.scalar.CopyFrom(scalar_to_protobuf(value))
    else:
        raise TypeError(
            f"Unsupported condition value type: {type(value)}. Please consult the documentation."
        )


def populate_publish_condition_batch_request_values(
    publish_request: PublishConditionBatchRequest, values: object
) -> None:
    """Assign a value to the scalar_values vector member of PublishConditionBatchRequest."""
    if isinstance(values, Vector):
        publish_request.scalar_values.CopyFrom(vector_to_protobuf(values))
    elif isinstance(values, Iterable):
        if not values:
            raise ValueError("Cannot publish an empty Iterable.")
        try:
            vector = Vector(values)
        except (TypeError, ValueError):
            raise TypeError(
                f"Unsupported iterable: {values}. Subtype must be bool, float, int, or string."
            )

        publish_request.scalar_values.CopyFrom(vector_to_protobuf(vector))
    else:
        raise TypeError(
            f"Unsupported condition values type: {type(values)}. Please consult the documentation."
        )


def populate_publish_measurement_request_value(
    publish_request: PublishMeasurementRequest, value: object
) -> None:
    """Assign a value to the appropriate field of a PublishMeasurementRequest object."""
    if isinstance(value, bool):
        publish_request.scalar.bool_value = value
    elif isinstance(value, int):
        publish_request.scalar.sint32_value = value
    elif isinstance(value, float):
        publish_request.scalar.double_value = value
    elif isinstance(value, str):
        publish_request.scalar.string_value = value
    elif isinstance(value, Scalar):
        publish_request.scalar.CopyFrom(scalar_to_protobuf(value))
    elif isinstance(value, Vector):
        publish_request.vector.CopyFrom(vector_to_protobuf(value))
    elif isinstance(value, AnalogWaveform):
        if value.dtype == np.float64:
            publish_request.double_analog_waveform.CopyFrom(
                float64_analog_waveform_to_protobuf(value)
            )
        elif value.dtype == np.int16:
            publish_request.i16_analog_waveform.CopyFrom(int16_analog_waveform_to_protobuf(value))
        else:
            raise TypeError(f"Unsupported AnalogWaveform dtype: {value.dtype}")
    elif isinstance(value, ComplexWaveform):
        if value.dtype == np.complex128:
            publish_request.double_complex_waveform.CopyFrom(
                float64_complex_waveform_to_protobuf(value)
            )
        elif value.dtype == ComplexInt32DType:
            publish_request.i16_complex_waveform.CopyFrom(int16_complex_waveform_to_protobuf(value))
        else:
            raise TypeError(f"Unsupported ComplexWaveform dtype: {value.dtype}")
    elif isinstance(value, Spectrum):
        if value.dtype == np.float64:
            publish_request.double_spectrum.CopyFrom(float64_spectrum_to_protobuf(value))
        else:
            raise TypeError(f"Unsupported Spectrum dtype: {value.dtype}")
    elif isinstance(value, DigitalWaveform):
        publish_request.digital_waveform.CopyFrom(digital_waveform_to_protobuf(value))
    elif isinstance(value, XYData):
        if value.dtype == np.float64:
            publish_request.x_y_data.CopyFrom(float64_xydata_to_protobuf(value))
        else:
            raise TypeError(f"Unsupported XYData dtype: {value.dtype}")
    elif isinstance(value, Iterable):
        if not value:
            raise ValueError("Cannot publish an empty Iterable.")
        try:
            vector = Vector(value)
        except (TypeError, ValueError):
            raise TypeError(
                f"Unsupported iterable: {value}. Subtype must be bool, float, int, or string."
            )

        publish_request.vector.CopyFrom(vector_to_protobuf(vector))
    else:
        raise TypeError(
            f"Unsupported measurement value type: {type(value)}. Please consult the documentation."
        )


def populate_publish_measurement_batch_request_values(
    publish_request: PublishMeasurementBatchRequest, values: object
) -> None:
    """Assign a value to the appropriate field of the PublishMeasurementBatchRequest object."""
    if isinstance(values, Vector):
        publish_request.scalar_values.CopyFrom(vector_to_protobuf(values))
    elif isinstance(values, Iterable):
        if not values:
            raise ValueError("Cannot publish an empty Iterable.")
        try:
            vector = Vector(values)
        except (TypeError, ValueError):
            raise TypeError(
                f"Unsupported iterable: {values}. Subtype must be bool, float, int, or string."
            )

        publish_request.scalar_values.CopyFrom(vector_to_protobuf(vector))
    else:
        raise TypeError(
            f"Unsupported measurement values type: {type(values)}. Please consult the documentation."
        )


def unpack_and_convert_from_protobuf_any(read_value: Any) -> object:
    """Convert from a packed pb.Any to the appropriate python object."""
    value_type = read_value.TypeName()
    if value_type == DoubleAnalogWaveform.DESCRIPTOR.full_name:
        double_analog_waveform = DoubleAnalogWaveform()
        read_value.Unpack(double_analog_waveform)
        return float64_analog_waveform_from_protobuf(double_analog_waveform)
    elif value_type == I16AnalogWaveform.DESCRIPTOR.full_name:
        i16_analog_waveform = I16AnalogWaveform()
        read_value.Unpack(i16_analog_waveform)
        return int16_analog_waveform_from_protobuf(i16_analog_waveform)
    elif value_type == DoubleComplexWaveform.DESCRIPTOR.full_name:
        double_complex_waveform = DoubleComplexWaveform()
        read_value.Unpack(double_complex_waveform)
        return float64_complex_waveform_from_protobuf(double_complex_waveform)
    elif value_type == I16ComplexWaveform.DESCRIPTOR.full_name:
        i16_complex_waveform = I16ComplexWaveform()
        read_value.Unpack(i16_complex_waveform)
        return int16_complex_waveform_from_protobuf(i16_complex_waveform)
    elif value_type == DoubleSpectrum.DESCRIPTOR.full_name:
        spectrum = DoubleSpectrum()
        read_value.Unpack(spectrum)
        return float64_spectrum_from_protobuf(spectrum)
    elif value_type == DigitalWaveformProto.DESCRIPTOR.full_name:
        digital_waveform = DigitalWaveformProto()
        read_value.Unpack(digital_waveform)
        return digital_waveform_from_protobuf(digital_waveform)
    elif value_type == DoubleXYData.DESCRIPTOR.full_name:
        xydata = DoubleXYData()
        read_value.Unpack(xydata)
        return float64_xydata_from_protobuf(xydata)
    elif value_type == VectorProto.DESCRIPTOR.full_name:
        vector = VectorProto()
        read_value.Unpack(vector)
        return vector_from_protobuf(vector)
    else:
        raise TypeError(f"Unsupported data type Name: {value_type}")


def get_publish_measurement_timestamp(
    publish_request: PublishMeasurementRequest, client_provided_timestamp: ht.datetime | None
) -> PrecisionTimestamp:
    """Determine the correct timestamp to use for publishing a measurement."""
    no_client_timestamp_provided = client_provided_timestamp is None
    if no_client_timestamp_provided:
        publish_time = hightime_datetime_to_protobuf(ht.datetime.now(std_datetime.timezone.utc))
    else:
        publish_time = hightime_datetime_to_protobuf(cast(ht.datetime, client_provided_timestamp))

    waveform_t0: PrecisionTimestamp | None = None
    value_case = publish_request.WhichOneof("value")
    if value_case == "double_analog_waveform":
        waveform_t0 = publish_request.double_analog_waveform.t0
    elif value_case == "i16_analog_waveform":
        waveform_t0 = publish_request.i16_analog_waveform.t0
    elif value_case == "double_complex_waveform":
        waveform_t0 = publish_request.double_complex_waveform.t0
    elif value_case == "i16_complex_waveform":
        waveform_t0 = publish_request.i16_complex_waveform.t0
    elif value_case == "digital_waveform":
        waveform_t0 = publish_request.digital_waveform.t0

    # If an initialized waveform t0 value is present
    if waveform_t0 is not None and waveform_t0 != PrecisionTimestamp():
        if no_client_timestamp_provided:
            # If the client did not provide a timestamp, use the waveform t0 value
            publish_time = waveform_t0
        elif publish_time != waveform_t0:
            raise ValueError(
                "The provided timestamp does not match the waveform t0. Please provide a matching timestamp or "
                "omit the timestamp to use the waveform t0."
            )
    return publish_time



"""Public API for accessing the NI Metadata Store."""

from ni.datastore.metadata._metadata_store_client import MetadataStoreClient
from ni.datastore.metadata._types._alias import Alias
from ni.datastore.metadata._types._alias_target_type import AliasTargetType
from ni.datastore.metadata._types._extension_schema import ExtensionSchema
from ni.datastore.metadata._types._hardware_item import HardwareItem
from ni.datastore.metadata._types._metadata_items import MetadataItems
from ni.datastore.metadata._types._operator import Operator
from ni.datastore.metadata._types._software_item import SoftwareItem
from ni.datastore.metadata._types._test import Test
from ni.datastore.metadata._types._test_adapter import TestAdapter
from ni.datastore.metadata._types._test_description import TestDescription
from ni.datastore.metadata._types._test_station import TestStation
from ni.datastore.metadata._types._uut import Uut
from ni.datastore.metadata._types._uut_instance import UutInstance

__all__ = [
    "Alias",
    "AliasTargetType",
    "ExtensionSchema",
    "HardwareItem",
    "MetadataItems",
    "MetadataStoreClient",
    "Operator",
    "SoftwareItem",
    "Test",
    "TestAdapter",
    "TestDescription",
    "TestStation",
    "Uut",
    "UutInstance",
]

# Hide that it was not defined in this top-level package
Alias.__module__ = __name__
AliasTargetType.__module__ = __name__
ExtensionSchema.__module__ = __name__
HardwareItem.__module__ = __name__
MetadataItems.__module__ = __name__
MetadataStoreClient.__module__ = __name__
Operator.__module__ = __name__
SoftwareItem.__module__ = __name__
Test.__module__ = __name__
TestAdapter.__module__ = __name__
TestDescription.__module__ = __name__
TestStation.__module__ = __name__
Uut.__module__ = __name__
UutInstance.__module__ = __name__



"""UUT data type for the Data Store Client."""

from __future__ import annotations

from typing import Iterable, Mapping, MutableMapping, MutableSequence

from ni.datastore.metadata._grpc_conversion import (
    populate_extension_value_message_map,
    populate_from_extension_value_message_map,
)
from ni.measurements.metadata.v1.metadata_store_pb2 import (
    Uut as UutProto,
)


class Uut:
    """Represents the metadata of a Unit Under Test (UUT).

    A UUT represents a model or type of device that can be tested, containing
    information like model name, family, manufacturers, and part number that
    describe the UUT type rather than specific instances.


    """

    __slots__ = (
        "_id",
        "model_name",
        "family",
        "_manufacturers",
        "part_number",
        "link",
        "_extension",
        "schema_id",
    )

    @property
    def manufacturers(self) -> MutableSequence[str]:
        """The manufacturers of the UUT."""
        return self._manufacturers

    @property
    def extension(self) -> MutableMapping[str, str]:
        """The extension of the UUT."""
        return self._extension

    @property
    def id(self) -> str:
        """The string id of the uut."""
        return self._id

    def __init__(
        self,
        *,
        model_name: str = "",
        family: str = "",
        manufacturers: Iterable[str] | None = None,
        part_number: str = "",
        link: str = "",
        extension: Mapping[str, str] | None = None,
        schema_id: str = "",
    ) -> None:
        """Initialize a Uut instance.

        Args:
            model_name: The name of the UUT model.
            family: The UUT family.
            manufacturers: List of manufacturers of the UUT.
            part_number: The part number of the UUT.
            link: A link to a resource that describes the UUT. This value is
                expected to be a valid URI.
            extension: Any extensions to be associated with the UUT.
            schema_id: The unique identifier of the schema that applies to this
                instance's extension. If any extension is associated with this
                instance, a schema_id must be provided, unless the UUT is
                created within the context of a test result, in which case the
                test result must have a schema_id.
        """
        self._id = ""
        self.model_name = model_name
        self.family = family
        self._manufacturers: MutableSequence[str] = (
            list(manufacturers) if manufacturers is not None else []
        )
        self.part_number = part_number
        self.link = link
        self._extension: MutableMapping[str, str] = dict(extension) if extension is not None else {}
        self.schema_id = schema_id

    @staticmethod
    def from_protobuf(uut_proto: UutProto) -> "Uut":
        """Create a Uut instance from a protobuf Uut message."""
        uut = Uut(
            model_name=uut_proto.model_name,
            family=uut_proto.family,
            manufacturers=uut_proto.manufacturers,
            part_number=uut_proto.part_number,
            link=uut_proto.link,
            schema_id=uut_proto.schema_id,
        )
        populate_from_extension_value_message_map(uut.extension, uut_proto.extension)
        uut._id = uut_proto.id
        return uut

    def to_protobuf(self) -> UutProto:
        """Convert this Uut to a protobuf Uut message."""
        uut_proto = UutProto(
            id=self.id,
            model_name=self.model_name,
            family=self.family,
            manufacturers=self.manufacturers,
            part_number=self.part_number,
            link=self.link,
            schema_id=self.schema_id,
        )
        populate_extension_value_message_map(uut_proto.extension, self.extension)
        return uut_proto

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, Uut):
            return NotImplemented
        return (
            self.id == other.id
            and self.model_name == other.model_name
            and self.family == other.family
            and self.manufacturers == other.manufacturers
            and self.part_number == other.part_number
            and self.link == other.link
            and self.extension == other.extension
            and self.schema_id == other.schema_id
        )

    def __str__(self) -> str:
        """Return a string representation of the Uut."""
        return str(self.to_protobuf())



"""Outcome data type for the Data Store Client."""

from __future__ import annotations

from enum import IntEnum

from ni.measurements.data.v1.data_store_pb2 import Outcome as OutcomeProto


class Outcome(IntEnum):
    """Represents the outcome of a measurement or test operation.

    The Outcome enum indicates whether a measurement or test passed, failed,
    or had an indeterminate result.
    """

    UNSPECIFIED = OutcomeProto.OUTCOME_UNSPECIFIED
    """The outcome is not specified or unknown."""

    PASSED = OutcomeProto.OUTCOME_PASSED
    """The measurement or test passed successfully."""

    FAILED = OutcomeProto.OUTCOME_FAILED
    """The measurement or test failed."""

    INDETERMINATE = OutcomeProto.OUTCOME_INDETERMINATE
    """The measurement or test result is indeterminate or inconclusive."""

    @classmethod
    def from_protobuf(cls, outcome_proto: OutcomeProto.ValueType) -> "Outcome":
        """Create an Outcome instance from a protobuf Outcome value.

        Args:
            outcome_proto: The protobuf Outcome value.

        Returns:
            The corresponding Outcome enum value.

        Raises:
            ValueError: If the protobuf value doesn't correspond to a known Outcome.
        """
        try:
            return cls(outcome_proto)
        except ValueError as e:
            raise ValueError(f"Unknown outcome value: {outcome_proto}") from e

    def to_protobuf(self) -> OutcomeProto.ValueType:
        """Convert this Outcome instance to a protobuf Outcome value.

        Returns:
            The corresponding protobuf Outcome value.
        """
        return OutcomeProto.ValueType(self.value)



"""Moniker data type for the Data Store Client."""

from __future__ import annotations

from ni.datamonikers.v1.data_moniker_pb2 import Moniker as MonikerProto


class Moniker:
    """Represents a data moniker for retrieving published data.

    A moniker provides the necessary information to locate and retrieve data
    from the data store, including the service location, data source, and
    data instance identifiers.
    """

    __slots__ = (
        "service_location",
        "data_source",
        "data_instance",
    )

    def __init__(
        self,
        *,
        service_location: str = "",
        data_source: str = "",
        data_instance: int = 0,
    ) -> None:
        """Initialize a Moniker instance.

        Args:
            service_location: The location of the service that stores the data.
            data_source: The identifier for the data source.
            data_instance: The instance number of the data.
        """
        self.service_location = service_location
        self.data_source = data_source
        self.data_instance = data_instance

    @staticmethod
    def from_protobuf(moniker_proto: MonikerProto) -> "Moniker":
        """Create a Moniker instance from a protobuf Moniker message."""
        return Moniker(
            service_location=moniker_proto.service_location,
            data_source=moniker_proto.data_source,
            data_instance=moniker_proto.data_instance,
        )

    def to_protobuf(self) -> MonikerProto:
        """Convert this Moniker instance to a protobuf Moniker message."""
        return MonikerProto(
            service_location=self.service_location,
            data_source=self.data_source,
            data_instance=self.data_instance,
        )

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, Moniker):
            return NotImplemented
        return (
            self.service_location == other.service_location
            and self.data_source == other.data_source
            and self.data_instance == other.data_instance
        )

    def __str__(self) -> str:
        """Return a string representation of the Moniker."""
        return str(self.to_protobuf())



"""Step data type for the Data Store Client."""

from __future__ import annotations

from typing import Mapping, MutableMapping

import hightime as ht
from ni.datastore.data._types._error_information import ErrorInformation
from ni.datastore.data._types._outcome import Outcome
from ni.datastore.metadata._grpc_conversion import (
    populate_extension_value_message_map,
    populate_from_extension_value_message_map,
)
from ni.measurements.data.v1.data_store_pb2 import Step as StepProto
from ni.protobuf.types.precision_timestamp_conversion import (
    hightime_datetime_from_protobuf,
    hightime_datetime_to_protobuf,
)


class Step:
    """Information about a step into which measurements and conditions are published.

    Represents a hierarchical execution step within a test result that can
    contain measurements and conditions. Steps are linked to a test result and
    can be organized into hierarchical structures using parent_step_id. Each
    step has test execution time, metadata, and optional extensions for custom
    metadata.
    """

    __slots__ = (
        "id",
        "parent_step_id",
        "test_result_id",
        "test_id",
        "name",
        "step_type",
        "notes",
        "start_date_time",
        "end_date_time",
        "link",
        "_extension",
        "schema_id",
        "error_information",
        "outcome",
    )

    @property
    def extension(self) -> MutableMapping[str, str]:
        """The extension of the step."""
        return self._extension

    def __init__(
        self,
        *,
        id: str = "",
        parent_step_id: str = "",
        test_result_id: str = "",
        test_id: str = "",
        name: str = "",
        step_type: str = "",
        notes: str = "",
        start_date_time: ht.datetime | None = None,
        end_date_time: ht.datetime | None = None,
        link: str = "",
        extension: Mapping[str, str] | None = None,
        schema_id: str = "",
        error_information: ErrorInformation | None = None,
        outcome: Outcome = Outcome.UNSPECIFIED,
    ) -> None:
        """Initialize a Step instance.

        Args:
            id: Unique identifier for the step.
            parent_step_id: ID of the parent step if this is a nested step.
            test_result_id: ID of the test result this step belongs to.
            test_id: ID of the test associated with this step.
            name: Human-readable name of the step.
            step_type: Type or category of the step.
            notes: Additional notes or comments about the step.
            start_date_time: The start date and time of the step execution.
            end_date_time: The end date and time of the step execution.
            link: Optional link to external resources for this step.
            extension: Additional extension attributes as key-value pairs.
            schema_id: ID of the extension schema for validating extensions.
            error_information: Error or exception information in case of
                step failure.
            outcome: The outcome of the step (PASSED, FAILED,
                INDETERMINATE, or UNSPECIFIED).
        """
        self.id = id
        self.parent_step_id = parent_step_id
        self.test_result_id = test_result_id
        self.test_id = test_id
        self.name = name
        self.step_type = step_type
        self.notes = notes
        self.start_date_time = start_date_time
        self.end_date_time = end_date_time
        self.link = link
        self._extension: MutableMapping[str, str] = dict(extension) if extension is not None else {}
        self.schema_id = schema_id
        self.error_information = error_information
        self.outcome = outcome

    @staticmethod
    def from_protobuf(step_proto: StepProto) -> "Step":
        """Create a Step instance from a protobuf Step message."""
        step = Step(
            id=step_proto.id,
            parent_step_id=step_proto.parent_step_id,
            test_result_id=step_proto.test_result_id,
            test_id=step_proto.test_id,
            name=step_proto.name,
            step_type=step_proto.step_type,
            notes=step_proto.notes,
            start_date_time=(
                hightime_datetime_from_protobuf(step_proto.start_date_time)
                if step_proto.HasField("start_date_time")
                else None
            ),
            end_date_time=(
                hightime_datetime_from_protobuf(step_proto.end_date_time)
                if step_proto.HasField("end_date_time")
                else None
            ),
            link=step_proto.link,
            schema_id=step_proto.schema_id,
            error_information=(
                ErrorInformation.from_protobuf(step_proto.error_information)
                if step_proto.HasField("error_information")
                else None
            ),
            outcome=Outcome.from_protobuf(step_proto.outcome),
        )
        populate_from_extension_value_message_map(step.extension, step_proto.extension)
        return step

    def to_protobuf(self) -> StepProto:
        """Convert this Step to a protobuf Step message."""
        step_proto = StepProto(
            id=self.id,
            parent_step_id=self.parent_step_id,
            test_result_id=self.test_result_id,
            test_id=self.test_id,
            name=self.name,
            step_type=self.step_type,
            notes=self.notes,
            start_date_time=(
                hightime_datetime_to_protobuf(self.start_date_time)
                if self.start_date_time
                else None
            ),
            end_date_time=(
                hightime_datetime_to_protobuf(self.end_date_time) if self.end_date_time else None
            ),
            link=self.link,
            schema_id=self.schema_id,
            error_information=(
                self.error_information.to_protobuf() if self.error_information is not None else None
            ),
            outcome=self.outcome.to_protobuf(),
        )
        populate_extension_value_message_map(step_proto.extension, self.extension)
        return step_proto

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, Step):
            return NotImplemented
        return (
            self.id == other.id
            and self.parent_step_id == other.parent_step_id
            and self.test_result_id == other.test_result_id
            and self.test_id == other.test_id
            and self.name == other.name
            and self.step_type == other.step_type
            and self.notes == other.notes
            and self.start_date_time == other.start_date_time
            and self.end_date_time == other.end_date_time
            and self.link == other.link
            and self.extension == other.extension
            and self.schema_id == other.schema_id
            and self.error_information == other.error_information
            and self.outcome == other.outcome
        )

    def __str__(self) -> str:
        """Return a string representation of the Step."""
        return str(self.to_protobuf())



"""MetadataItems data type for the Data Store Client."""

from __future__ import annotations

from collections.abc import Sequence

from ni.datastore.metadata._types._hardware_item import HardwareItem
from ni.datastore.metadata._types._operator import Operator
from ni.datastore.metadata._types._software_item import SoftwareItem
from ni.datastore.metadata._types._test import Test
from ni.datastore.metadata._types._test_adapter import TestAdapter
from ni.datastore.metadata._types._test_description import TestDescription
from ni.datastore.metadata._types._test_station import TestStation
from ni.datastore.metadata._types._uut import Uut
from ni.datastore.metadata._types._uut_instance import UutInstance
from ni.measurements.metadata.v1.metadata_store_service_pb2 import (
    CreateFromJsonDocumentResponse,
)


class MetadataItems:
    """Represents a collection of metadata items created from a JSON document.

    This class contains the results of creating metadata items from a JSON document,
    organizing them by type for easy access.
    """

    __slots__ = (
        "uut_instances",
        "uuts",
        "operators",
        "test_descriptions",
        "tests",
        "test_stations",
        "hardware_items",
        "software_items",
        "test_adapters",
    )

    def __init__(
        self,
        *,
        uut_instances: Sequence[UutInstance] = (),
        uuts: Sequence[Uut] = (),
        operators: Sequence[Operator] = (),
        test_descriptions: Sequence[TestDescription] = (),
        tests: Sequence[Test] = (),
        test_stations: Sequence[TestStation] = (),
        hardware_items: Sequence[HardwareItem] = (),
        software_items: Sequence[SoftwareItem] = (),
        test_adapters: Sequence[TestAdapter] = (),
    ) -> None:
        """Initialize a MetadataItems instance.

        Args:
            uut_instances: A sequence of UUT instances created from the JSON document.
            uuts: A sequence of UUTs created from the JSON document.
            operators: A sequence of operators created from the JSON document.
            test_descriptions: A sequence of test descriptions created from the JSON document.
            tests: A sequence of tests created from the JSON document.
            test_stations: A sequence of test stations created from the JSON document.
            hardware_items: A sequence of hardware items created from the JSON document.
            software_items: A sequence of software items created from the JSON document.
            test_adapters: A sequence of test adapters created from the JSON document.
        """
        self.uut_instances = uut_instances
        self.uuts = uuts
        self.operators = operators
        self.test_descriptions = test_descriptions
        self.tests = tests
        self.test_stations = test_stations
        self.hardware_items = hardware_items
        self.software_items = software_items
        self.test_adapters = test_adapters

    @staticmethod
    def from_protobuf(response: CreateFromJsonDocumentResponse) -> "MetadataItems":
        """Create a MetadataItems instance from a protobuf CreateFromJsonDocumentResponse.

        Args:
            response: The protobuf response containing the created metadata items.

        Returns:
            MetadataItems: A new MetadataItems instance with all the created items.
        """
        return MetadataItems(
            uut_instances=[
                UutInstance.from_protobuf(uut_instance) for uut_instance in response.uut_instances
            ],
            uuts=[Uut.from_protobuf(uut) for uut in response.uuts],
            operators=[Operator.from_protobuf(operator) for operator in response.operators],
            test_descriptions=[
                TestDescription.from_protobuf(test_description)
                for test_description in response.test_descriptions
            ],
            tests=[Test.from_protobuf(test) for test in response.tests],
            test_stations=[
                TestStation.from_protobuf(test_station) for test_station in response.test_stations
            ],
            hardware_items=[
                HardwareItem.from_protobuf(hardware_item)
                for hardware_item in response.hardware_items
            ],
            software_items=[
                SoftwareItem.from_protobuf(software_item)
                for software_item in response.software_items
            ],
            test_adapters=[
                TestAdapter.from_protobuf(test_adapter) for test_adapter in response.test_adapters
            ],
        )

    def __eq__(self, other: object) -> bool:
        """Determine equality between MetadataItems instances."""
        if not isinstance(other, MetadataItems):
            return False

        return (
            list(self.uut_instances) == list(other.uut_instances)
            and list(self.uuts) == list(other.uuts)
            and list(self.operators) == list(other.operators)
            and list(self.test_descriptions) == list(other.test_descriptions)
            and list(self.tests) == list(other.tests)
            and list(self.test_stations) == list(other.test_stations)
            and list(self.hardware_items) == list(other.hardware_items)
            and list(self.software_items) == list(other.software_items)
            and list(self.test_adapters) == list(other.test_adapters)
        )

    def __str__(self) -> str:
        """Return a string representation of the MetadataItems."""
        counts = []
        if self.uut_instances:
            counts.append(f"{len(self.uut_instances)} UUT instances")
        if self.uuts:
            counts.append(f"{len(self.uuts)} UUTs")
        if self.operators:
            counts.append(f"{len(self.operators)} operators")
        if self.test_descriptions:
            counts.append(f"{len(self.test_descriptions)} test descriptions")
        if self.tests:
            counts.append(f"{len(self.tests)} tests")
        if self.test_stations:
            counts.append(f"{len(self.test_stations)} test stations")
        if self.hardware_items:
            counts.append(f"{len(self.hardware_items)} hardware items")
        if self.software_items:
            counts.append(f"{len(self.software_items)} software items")
        if self.test_adapters:
            counts.append(f"{len(self.test_adapters)} test adapters")

        if counts:
            return f"MetadataItems({', '.join(counts)})"
        else:
            return "MetadataItems(empty)"



"""Operator data type for the Data Store Client."""

from __future__ import annotations

from typing import Mapping, MutableMapping

from ni.datastore.metadata._grpc_conversion import (
    populate_extension_value_message_map,
    populate_from_extension_value_message_map,
)
from ni.measurements.metadata.v1.metadata_store_pb2 import (
    Operator as OperatorProto,
)


class Operator:
    """Represents the metadata of the operator that took the test step.

    An operator contains information about the person or entity responsible for
    conducting tests, including their name and role.
    """

    __slots__ = (
        "_id",
        "name",
        "role",
        "link",
        "_extension",
        "schema_id",
    )

    @property
    def extension(self) -> MutableMapping[str, str]:
        """The extension of the operator."""
        return self._extension

    @property
    def id(self) -> str:
        """The string id of the operator."""
        return self._id

    def __init__(
        self,
        *,
        name: str = "",
        role: str = "",
        link: str = "",
        extension: Mapping[str, str] | None = None,
        schema_id: str = "",
    ) -> None:
        """Initialize an Operator instance.

        Args:
            name: The name of the operator.
            role: The role of the operator.
            link: A link to a resource that describes the operator. This value
                is expected to be a valid URI.
            extension: Any extensions to be associated with the operator.
            schema_id: The unique identifier of the schema that applies to this
                instance's extension. If any extension is associated with this
                instance, a schema_id must be provided, unless the operator is
                created within the context of a test result, in which case the
                test result must have a schema_id.
        """
        self._id = ""
        self.name = name
        self.role = role
        self.link = link
        self._extension: MutableMapping[str, str] = dict(extension) if extension is not None else {}
        self.schema_id = schema_id

    @staticmethod
    def from_protobuf(operator_proto: OperatorProto) -> "Operator":
        """Create an Operator instance from a protobuf Operator message."""
        operator = Operator(
            name=operator_proto.name,
            role=operator_proto.role,
            link=operator_proto.link,
            schema_id=operator_proto.schema_id,
        )
        populate_from_extension_value_message_map(operator.extension, operator_proto.extension)
        operator._id = operator_proto.id
        return operator

    def to_protobuf(self) -> OperatorProto:
        """Convert this Operator to a protobuf Operator message."""
        operator_proto = OperatorProto(
            id=self.id,
            name=self.name,
            role=self.role,
            link=self.link,
            schema_id=self.schema_id,
        )
        populate_extension_value_message_map(operator_proto.extension, self.extension)
        return operator_proto

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, Operator):
            return NotImplemented
        return (
            self.id == other.id
            and self.name == other.name
            and self.role == other.role
            and self.link == other.link
            and self.extension == other.extension
            and self.schema_id == other.schema_id
        )

    def __str__(self) -> str:
        """Return a string representation of the Operator."""
        return str(self.to_protobuf())



"""Public API for accessing the NI Data Store."""

from ni.datastore.data._data_store_client import DataStoreClient
from ni.datastore.data._types._error_information import ErrorInformation
from ni.datastore.data._types._moniker import Moniker
from ni.datastore.data._types._outcome import Outcome
from ni.datastore.data._types._published_condition import PublishedCondition
from ni.datastore.data._types._published_measurement import PublishedMeasurement
from ni.datastore.data._types._step import Step
from ni.datastore.data._types._test_result import TestResult

__all__ = [
    "DataStoreClient",
    "ErrorInformation",
    "Moniker",
    "Outcome",
    "PublishedCondition",
    "PublishedMeasurement",
    "Step",
    "TestResult",
]

# Hide that it was not defined in this top-level package
DataStoreClient.__module__ = __name__
ErrorInformation.__module__ = __name__
Moniker.__module__ = __name__
Outcome.__module__ = __name__
PublishedCondition.__module__ = __name__
PublishedMeasurement.__module__ = __name__
Step.__module__ = __name__
TestResult.__module__ = __name__



"""Test Result data type for the Data Store Client."""

from __future__ import annotations

from typing import Iterable, Mapping, MutableMapping, MutableSequence

import hightime as ht
from ni.datastore.data._types._error_information import ErrorInformation
from ni.datastore.data._types._outcome import Outcome
from ni.datastore.metadata._grpc_conversion import (
    populate_extension_value_message_map,
    populate_from_extension_value_message_map,
)
from ni.measurements.data.v1.data_store_pb2 import TestResult as TestResultProto
from ni.protobuf.types.precision_timestamp_conversion import (
    hightime_datetime_from_protobuf,
    hightime_datetime_to_protobuf,
)


class TestResult:
    """Information about a test result.

    Represents the results of executing a test sequence, including metadata
    about the execution environment (UUT, operator, test station), test
    execution time, outcome, and associations with hardware/software components
    and test adapters. Each step which references measurement data and
    conditions is associated with a test result.
    """

    __slots__ = (
        "id",
        "uut_instance_id",
        "operator_id",
        "test_station_id",
        "test_description_id",
        "_software_item_ids",
        "_hardware_item_ids",
        "_test_adapter_ids",
        "name",
        "start_date_time",
        "end_date_time",
        "outcome",
        "link",
        "_extension",
        "schema_id",
        "error_information",
    )

    @property
    def software_item_ids(self) -> MutableSequence[str]:
        """The software item IDs associated with the test result."""
        return self._software_item_ids

    @property
    def hardware_item_ids(self) -> MutableSequence[str]:
        """The hardware item IDs associated with the test result."""
        return self._hardware_item_ids

    @property
    def test_adapter_ids(self) -> MutableSequence[str]:
        """The test adapter IDs associated with the test result."""
        return self._test_adapter_ids

    @property
    def extension(self) -> MutableMapping[str, str]:
        """The extension of the test result."""
        return self._extension

    def __init__(
        self,
        *,
        id: str = "",
        uut_instance_id: str = "",
        operator_id: str = "",
        test_station_id: str = "",
        test_description_id: str = "",
        software_item_ids: Iterable[str] | None = None,
        hardware_item_ids: Iterable[str] | None = None,
        test_adapter_ids: Iterable[str] | None = None,
        name: str = "",
        start_date_time: ht.datetime | None = None,
        end_date_time: ht.datetime | None = None,
        outcome: Outcome = Outcome.UNSPECIFIED,
        link: str = "",
        extension: Mapping[str, str] | None = None,
        schema_id: str = "",
        error_information: ErrorInformation | None = None,
    ) -> None:
        """Initialize a TestResult instance.

        Args:
            id: Unique identifier for the test result.
            uut_instance_id: ID of the UUT instance that was tested.
            operator_id: ID of the operator who ran the test.
            test_station_id: ID of the test station used.
            test_description_id: ID of the test description that was executed.
            software_item_ids: IDs of software items used in the test.
            hardware_item_ids: IDs of hardware items used in the test.
            test_adapter_ids: IDs of test adapters used in the test.
            name: Human-readable name for the test result.
            start_date_time: The start date and time of the test execution.
            end_date_time: The end date and time of the test execution.
            outcome: The outcome of the test execution (PASSED, FAILED,
                INDETERMINATE, or UNSPECIFIED).
            link: Optional link to external resources for this test result.
            extension: Additional extension attributes as key-value pairs.
            schema_id: ID of the extension schema for validating extensions.
            error_information: Error or exception information in case of
                test result failure.
        """
        self.id = id
        self.uut_instance_id = uut_instance_id
        self.operator_id = operator_id
        self.test_station_id = test_station_id
        self.test_description_id = test_description_id
        self._software_item_ids: MutableSequence[str] = (
            list(software_item_ids) if software_item_ids is not None else []
        )
        self._hardware_item_ids: MutableSequence[str] = (
            list(hardware_item_ids) if hardware_item_ids is not None else []
        )
        self._test_adapter_ids: MutableSequence[str] = (
            list(test_adapter_ids) if test_adapter_ids is not None else []
        )
        self.name = name
        self.start_date_time = start_date_time
        self.end_date_time = end_date_time
        self.outcome = outcome
        self.link = link
        self._extension: MutableMapping[str, str] = dict(extension) if extension is not None else {}
        self.schema_id = schema_id
        self.error_information = error_information

    @staticmethod
    def from_protobuf(test_result_proto: TestResultProto) -> "TestResult":
        """Create a TestResult instance from a protobuf TestResult message."""
        test_result = TestResult(
            id=test_result_proto.id,
            uut_instance_id=test_result_proto.uut_instance_id,
            operator_id=test_result_proto.operator_id,
            test_station_id=test_result_proto.test_station_id,
            test_description_id=test_result_proto.test_description_id,
            software_item_ids=test_result_proto.software_item_ids,
            hardware_item_ids=test_result_proto.hardware_item_ids,
            test_adapter_ids=test_result_proto.test_adapter_ids,
            name=test_result_proto.name,
            start_date_time=(
                hightime_datetime_from_protobuf(test_result_proto.start_date_time)
                if test_result_proto.HasField("start_date_time")
                else None
            ),
            end_date_time=(
                hightime_datetime_from_protobuf(test_result_proto.end_date_time)
                if test_result_proto.HasField("end_date_time")
                else None
            ),
            outcome=Outcome.from_protobuf(test_result_proto.outcome),
            link=test_result_proto.link,
            schema_id=test_result_proto.schema_id,
            error_information=(
                ErrorInformation.from_protobuf(test_result_proto.error_information)
                if test_result_proto.HasField("error_information")
                else None
            ),
        )
        populate_from_extension_value_message_map(
            test_result.extension, test_result_proto.extension
        )
        return test_result

    def to_protobuf(self) -> TestResultProto:
        """Convert this TestResult to a protobuf TestResult message."""
        test_result_proto = TestResultProto(
            id=self.id,
            uut_instance_id=self.uut_instance_id,
            operator_id=self.operator_id,
            test_station_id=self.test_station_id,
            test_description_id=self.test_description_id,
            software_item_ids=self.software_item_ids,
            hardware_item_ids=self.hardware_item_ids,
            test_adapter_ids=self.test_adapter_ids,
            name=self.name,
            start_date_time=(
                hightime_datetime_to_protobuf(self.start_date_time)
                if self.start_date_time
                else None
            ),
            end_date_time=(
                hightime_datetime_to_protobuf(self.end_date_time) if self.end_date_time else None
            ),
            outcome=self.outcome.to_protobuf(),
            link=self.link,
            schema_id=self.schema_id,
            error_information=(
                self.error_information.to_protobuf() if self.error_information is not None else None
            ),
        )
        populate_extension_value_message_map(test_result_proto.extension, self.extension)
        return test_result_proto

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, TestResult):
            return NotImplemented
        return (
            self.id == other.id
            and self.uut_instance_id == other.uut_instance_id
            and self.operator_id == other.operator_id
            and self.test_station_id == other.test_station_id
            and self.test_description_id == other.test_description_id
            and self.software_item_ids == other.software_item_ids
            and self.hardware_item_ids == other.hardware_item_ids
            and self.test_adapter_ids == other.test_adapter_ids
            and self.name == other.name
            and self.start_date_time == other.start_date_time
            and self.end_date_time == other.end_date_time
            and self.outcome == other.outcome
            and self.link == other.link
            and self.extension == other.extension
            and self.schema_id == other.schema_id
            and self.error_information == other.error_information
        )

    def __str__(self) -> str:
        """Return a string representation of the TestResult."""
        return str(self.to_protobuf())



"""Test Station data type for the Data Store Client."""

from __future__ import annotations

from typing import Mapping, MutableMapping

from ni.datastore.metadata._grpc_conversion import (
    populate_extension_value_message_map,
    populate_from_extension_value_message_map,
)
from ni.measurements.metadata.v1.metadata_store_pb2 import (
    TestStation as TestStationProto,
)


class TestStation:
    """Represents the metadata of a test station.

    A test station contains information about the physical location or setup
    where testing is performed, including its name and asset identifier for
    tracking and inventory purposes.


    """

    __slots__ = (
        "_id",
        "name",
        "asset_identifier",
        "link",
        "_extension",
        "schema_id",
    )

    @property
    def extension(self) -> MutableMapping[str, str]:
        """The extension of the test station."""
        return self._extension

    @property
    def id(self) -> str:
        """The string id of the test station."""
        return self._id

    def __init__(
        self,
        *,
        name: str = "",
        asset_identifier: str = "",
        link: str = "",
        extension: Mapping[str, str] | None = None,
        schema_id: str = "",
    ) -> None:
        """Initialize a TestStation instance.

        Args:
            name: The name of the test station.
            asset_identifier: The asset identifier of the test station.
            link: A link to a resource that describes the test station. This
                value is expected to be a valid URI.
            extension: Any extensions to be associated with the test station.
            schema_id: The unique identifier of the schema that applies to this
                instance's extension. If any extension is associated with this
                instance, a schema_id must be provided, unless the test station
                is created within the context of a test result, in which case
                the test result must have a schema_id.
        """
        self._id = ""
        self.name = name
        self.asset_identifier = asset_identifier
        self.link = link
        self._extension: MutableMapping[str, str] = dict(extension) if extension is not None else {}
        self.schema_id = schema_id

    @staticmethod
    def from_protobuf(test_station_proto: TestStationProto) -> "TestStation":
        """Create a TestStation instance from a protobuf TestStation message."""
        test_station = TestStation(
            name=test_station_proto.name,
            asset_identifier=test_station_proto.asset_identifier,
            link=test_station_proto.link,
            schema_id=test_station_proto.schema_id,
        )
        populate_from_extension_value_message_map(
            test_station.extension, test_station_proto.extension
        )
        test_station._id = test_station_proto.id
        return test_station

    def to_protobuf(self) -> TestStationProto:
        """Convert this TestStation to a protobuf TestStation message."""
        test_station_proto = TestStationProto(
            id=self.id,
            name=self.name,
            asset_identifier=self.asset_identifier,
            link=self.link,
            schema_id=self.schema_id,
        )
        populate_extension_value_message_map(test_station_proto.extension, self.extension)
        return test_station_proto

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, TestStation):
            return NotImplemented
        return (
            self.id == other.id
            and self.name == other.name
            and self.asset_identifier == other.asset_identifier
            and self.link == other.link
            and self.extension == other.extension
            and self.schema_id == other.schema_id
        )

    def __str__(self) -> str:
        """Return a string representation of the TestStation."""
        return str(self.to_protobuf())



"""AliasTargetType data type for the Data Store Client."""

from __future__ import annotations

from enum import IntEnum

from ni.measurements.metadata.v1.metadata_store_pb2 import (
    AliasTargetType as AliasTargetTypeProto,
)


class AliasTargetType(IntEnum):
    """Represents the type of target that an alias can reference.

    The AliasTargetType enum indicates what kind of metadata entity
    an alias is pointing to.
    """

    UNSPECIFIED = AliasTargetTypeProto.ALIAS_TARGET_TYPE_UNSPECIFIED
    """The alias target type is not specified or unknown."""

    UUT_INSTANCE = AliasTargetTypeProto.ALIAS_TARGET_TYPE_UUT_INSTANCE
    """The alias targets a UUT instance."""

    UUT = AliasTargetTypeProto.ALIAS_TARGET_TYPE_UUT
    """The alias targets a UUT."""

    HARDWARE_ITEM = AliasTargetTypeProto.ALIAS_TARGET_TYPE_HARDWARE_ITEM
    """The alias targets a hardware item."""

    SOFTWARE_ITEM = AliasTargetTypeProto.ALIAS_TARGET_TYPE_SOFTWARE_ITEM
    """The alias targets a software item."""

    OPERATOR = AliasTargetTypeProto.ALIAS_TARGET_TYPE_OPERATOR
    """The alias targets an operator."""

    TEST_DESCRIPTION = AliasTargetTypeProto.ALIAS_TARGET_TYPE_TEST_DESCRIPTION
    """The alias targets a test description."""

    TEST = AliasTargetTypeProto.ALIAS_TARGET_TYPE_TEST
    """The alias targets a test."""

    TEST_STATION = AliasTargetTypeProto.ALIAS_TARGET_TYPE_TEST_STATION
    """The alias targets a test station."""

    TEST_ADAPTER = AliasTargetTypeProto.ALIAS_TARGET_TYPE_TEST_ADAPTER
    """The alias targets a test adapter."""

    @classmethod
    def from_protobuf(
        cls, alias_target_type_proto: AliasTargetTypeProto.ValueType
    ) -> "AliasTargetType":
        """Create an AliasTargetType instance from a protobuf AliasTargetType value.

        Args:
            alias_target_type_proto: The protobuf AliasTargetType value.

        Returns:
            The corresponding AliasTargetType enum value.

        Raises:
            ValueError: If the protobuf value doesn't correspond to a known AliasTargetType.
        """
        try:
            return cls(alias_target_type_proto)
        except ValueError as e:
            raise ValueError(f"Unknown alias target type value: {alias_target_type_proto}") from e

    def to_protobuf(self) -> AliasTargetTypeProto.ValueType:
        """Convert this AliasTargetType instance to a protobuf AliasTargetType value.

        Returns:
            The corresponding protobuf AliasTargetType value.
        """
        return AliasTargetTypeProto.ValueType(self.value)



"""Software Item data type for the Data Store Client."""

from __future__ import annotations

from typing import Mapping, MutableMapping

from ni.datastore.metadata._grpc_conversion import (
    populate_extension_value_message_map,
    populate_from_extension_value_message_map,
)
from ni.measurements.metadata.v1.metadata_store_pb2 import (
    SoftwareItem as SoftwareItemProto,
)


class SoftwareItem:
    """Represents the metadata of the software item used to take measurements.

    A software item contains information about the software product and version
    used during testing.
    """

    __slots__ = (
        "_id",
        "product",
        "version",
        "link",
        "_extension",
        "schema_id",
    )

    @property
    def extension(self) -> MutableMapping[str, str]:
        """The extension of the software item."""
        return self._extension

    @property
    def id(self) -> str:
        """The string id of the software item."""
        return self._id

    def __init__(
        self,
        *,
        product: str = "",
        version: str = "",
        link: str = "",
        extension: Mapping[str, str] | None = None,
        schema_id: str = "",
    ) -> None:
        """Initialize a SoftwareItem instance.

        Args:
            product: The product name of the software item.
            version: The version of the software item.
            link: A link to a resource that describes the software item. This
                value is expected to be a valid URI.
            extension: Any extensions to be associated with the software item.
            schema_id: The unique identifier of the schema that applies to this
                instance's extension. If any extension is associated with this
                instance, a schema_id must be provided, unless the software item
                is created within the context of a test result, in which case
                the test result must have a schema_id.
        """
        self._id = ""
        self.product = product
        self.version = version
        self.link = link
        self._extension: MutableMapping[str, str] = dict(extension) if extension is not None else {}
        self.schema_id = schema_id

    @staticmethod
    def from_protobuf(software_item_proto: SoftwareItemProto) -> "SoftwareItem":
        """Create a SoftwareItem instance from a protobuf SoftwareItem message."""
        software_item = SoftwareItem(
            product=software_item_proto.product,
            version=software_item_proto.version,
            link=software_item_proto.link,
            schema_id=software_item_proto.schema_id,
        )
        populate_from_extension_value_message_map(
            software_item.extension, software_item_proto.extension
        )
        software_item._id = software_item_proto.id
        return software_item

    def to_protobuf(self) -> SoftwareItemProto:
        """Convert this SoftwareItem to a protobuf SoftwareItem message."""
        software_item_proto = SoftwareItemProto(
            id=self.id,
            product=self.product,
            version=self.version,
            link=self.link,
            schema_id=self.schema_id,
        )
        populate_extension_value_message_map(software_item_proto.extension, self.extension)
        return software_item_proto

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, SoftwareItem):
            return NotImplemented
        return (
            self.id == other.id
            and self.product == other.product
            and self.version == other.version
            and self.link == other.link
            and self.extension == other.extension
            and self.schema_id == other.schema_id
        )

    def __str__(self) -> str:
        """Return a string representation of the SoftwareItem."""
        return str(self.to_protobuf())



"""Hardware Item data type for the Data Store Client."""

from __future__ import annotations

from typing import Mapping, MutableMapping

from ni.datastore.metadata._grpc_conversion import (
    populate_extension_value_message_map,
    populate_from_extension_value_message_map,
)
from ni.measurements.metadata.v1.metadata_store_pb2 import (
    HardwareItem as HardwareItemProto,
)


class HardwareItem:
    """Represents the metadata of a hardware item used to take measurements.

    A hardware item contains information about physical test equipment
    including manufacturer, model, serial number, calibration dates, and asset
    tracking information.
    """

    __slots__ = (
        "_id",
        "manufacturer",
        "model",
        "serial_number",
        "part_number",
        "asset_identifier",
        "calibration_due_date",
        "link",
        "_extension",
        "schema_id",
    )

    @property
    def extension(self) -> MutableMapping[str, str]:
        """The extension of the hardware item."""
        return self._extension

    @property
    def id(self) -> str:
        """The string id of the hardware item."""
        return self._id

    def __init__(
        self,
        *,
        manufacturer: str = "",
        model: str = "",
        serial_number: str = "",
        part_number: str = "",
        asset_identifier: str = "",
        calibration_due_date: str = "",
        link: str = "",
        extension: Mapping[str, str] | None = None,
        schema_id: str = "",
    ) -> None:
        """Initialize a HardwareItem instance.

        Args:
            manufacturer: The manufacturer of the hardware item.
            model: The model of the hardware item.
            serial_number: The serial number of the hardware item.
            part_number: The part number of the hardware item.
            asset_identifier: The asset identifier of the hardware item.
            calibration_due_date: The calibration due date of the hardware item.
            link: A link to a resource that describes the hardware item. This
                value is expected to be a valid URI.
            extension: Any extensions to be associated with the hardware item.
            schema_id: The unique identifier of the schema that applies to this
                instance's extension. If any extension is associated with this
                instance, a schema_id must be provided, unless the hardware item
                is created within the context of a test result, in which case
                the test result must have a schema_id.
        """
        self._id = ""
        self.manufacturer = manufacturer
        self.model = model
        self.serial_number = serial_number
        self.part_number = part_number
        self.asset_identifier = asset_identifier
        self.calibration_due_date = calibration_due_date
        self.link = link
        self._extension: MutableMapping[str, str] = dict(extension) if extension is not None else {}
        self.schema_id = schema_id

    @staticmethod
    def from_protobuf(hardware_item_proto: HardwareItemProto) -> "HardwareItem":
        """Create a HardwareItem instance from a protobuf HardwareItem message."""
        hardware_item = HardwareItem(
            manufacturer=hardware_item_proto.manufacturer,
            model=hardware_item_proto.model,
            serial_number=hardware_item_proto.serial_number,
            part_number=hardware_item_proto.part_number,
            asset_identifier=hardware_item_proto.asset_identifier,
            calibration_due_date=hardware_item_proto.calibration_due_date,
            link=hardware_item_proto.link,
            schema_id=hardware_item_proto.schema_id,
        )
        populate_from_extension_value_message_map(
            hardware_item.extension, hardware_item_proto.extension
        )
        hardware_item._id = hardware_item_proto.id
        return hardware_item

    def to_protobuf(self) -> HardwareItemProto:
        """Convert this HardwareItem to a protobuf HardwareItem message."""
        hardware_item_proto = HardwareItemProto(
            id=self.id,
            manufacturer=self.manufacturer,
            model=self.model,
            serial_number=self.serial_number,
            part_number=self.part_number,
            asset_identifier=self.asset_identifier,
            calibration_due_date=self.calibration_due_date,
            link=self.link,
            schema_id=self.schema_id,
        )
        populate_extension_value_message_map(hardware_item_proto.extension, self.extension)
        return hardware_item_proto

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, HardwareItem):
            return NotImplemented
        return (
            self.id == other.id
            and self.manufacturer == other.manufacturer
            and self.model == other.model
            and self.serial_number == other.serial_number
            and self.part_number == other.part_number
            and self.asset_identifier == other.asset_identifier
            and self.calibration_due_date == other.calibration_due_date
            and self.link == other.link
            and self.extension == other.extension
            and self.schema_id == other.schema_id
        )

    def __str__(self) -> str:
        """Return a string representation of the HardwareItem."""
        return str(self.to_protobuf())



"""Error Information data type for the Data Store Client."""

from __future__ import annotations

from ni.measurements.data.v1.data_store_pb2 import (
    ErrorInformation as ErrorInformationProto,
)


class ErrorInformation:
    """Represents error or exception information in case of measurement failure.

    An ErrorInformation contains an error code, descriptive message, and
    source information to help identify and diagnose issues during measurement
    execution.
    """

    __slots__ = (
        "error_code",
        "message",
        "source",
    )

    def __init__(
        self,
        *,
        error_code: int = 0,
        message: str = "",
        source: str = "",
    ) -> None:
        """Initialize an ErrorInformation instance.

        Args:
            error_code: The numeric error code associated with the error.
            message: A descriptive message explaining the error.
            source: The source or location where the error occurred.
        """
        self.error_code = error_code
        self.message = message
        self.source = source

    @staticmethod
    def from_protobuf(
        error_information_proto: ErrorInformationProto,
    ) -> "ErrorInformation":
        """Create an ErrorInformation instance from a protobuf ErrorInformation message."""
        return ErrorInformation(
            error_code=error_information_proto.error_code,
            message=error_information_proto.message,
            source=error_information_proto.source,
        )

    def to_protobuf(self) -> ErrorInformationProto:
        """Convert this ErrorInformation instance to a protobuf ErrorInformation message."""
        return ErrorInformationProto(
            error_code=self.error_code,
            message=self.message,
            source=self.source,
        )

    def __eq__(self, other: object) -> bool:
        """Determine equality."""
        if not isinstance(other, ErrorInformation):
            return NotImplemented
        return (
            self.error_code == other.error_code
            and self.message == other.message
            and self.source == other.source
        )

    def __str__(self) -> str:
        """Return a string representation of the ErrorInformation."""
        return str(self.to_protobuf())



"""Helper methods to convert between python and protobuf types."""

from __future__ import annotations

import logging
from typing import MutableMapping

from google.protobuf.internal.containers import MessageMap
from ni.measurements.metadata.v1.metadata_store_pb2 import ExtensionValue

_logger = logging.getLogger(__name__)


def populate_extension_value_message_map(
    destination: MessageMap[str, ExtensionValue],
    source: MutableMapping[str, str],
) -> None:
    """Populate a gRPC message map of string keys to ExtensionValue.

    The input is a mapping of string keys to string values.
    """
    for key, value in source.items():
        destination[key].string_value = value


def populate_from_extension_value_message_map(
    destination: MutableMapping[str, str],
    source: MessageMap[str, ExtensionValue],
) -> None:
    """Populate a mapping of string keys to stringvalues.

    The input is a gRPC message map of string keys to ExtensionValue.
    """
    for key, extension_value in source.items():
        value_case = extension_value.WhichOneof("metadata")
        if value_case == "string_value":
            destination[key] = extension_value.string_value
        else:
            raise TypeError(f"Unsupported ExtensionValue type for key '{key}': {value_case}")
